import os
import warnings
from datetime import datetime
from unittest import skip

import pytest
import pytz
import requests

from pykechain.enums import (
    ActivityClassification,
    ActivityRootNames,
    ActivityStatus,
    ActivityType,
    Category,
    Multiplicity,
    NotificationEvent,
    PaperOrientation,
    PaperSize,
    PropertyType,
    activity_root_name_by_classification,
)
from pykechain.exceptions import (
    APIError,
    IllegalArgumentError,
    MultipleFoundError,
    NotFoundError,
)
from pykechain.models import Activity
from pykechain.models.representations import CustomIconRepresentation
from pykechain.utils import slugify_ref, temp_chdir
from tests.classes import TestBetamax

ISOFORMAT = "%Y-%m-%dT%H:%M:%SZ"
ISOFORMAT_HIGHPRECISION = "%Y-%m-%dT%H:%M:%S.%fZ"


class TestActivityConstruction(TestBetamax):
    def setUp(self):
        super().setUp()
        self.process = self.project.create_activity(
            name="__Test process",
            activity_type=ActivityType.PROCESS,
        )
        self.task = None

    def tearDown(self):
        for activity in [self.task, self.process]:
            if activity:
                try:
                    activity.delete()
                except APIError:
                    pass
        super().tearDown()

    def test_create_with_inputs(self):
        name = "__Testing task"
        ref = slugify_ref(name)
        description = "My new task"
        status = ActivityStatus.OPEN
        activity_type = ActivityType.TASK
        classification = ActivityClassification.WORKFLOW

        # setUp
        self.task = self.client.create_activity(
            name=name,
            ref=ref,
            parent=self.process,
            status=status,
            description=description,
            start_date=self.time,
            due_date=self.time,
            activity_type=activity_type,
            classification=classification,
            activity_options=dict(
                representations=[
                    CustomIconRepresentation(value="pennant").as_json(),
                ],
            ),
        )

        # testing
        self.assertIsInstance(self.task, Activity)
        self.assertEqual(name, self.task.name)
        self.assertEqual(ref, self.task.ref)
        self.assertEqual(status, self.task.status)
        self.assertTrue(description, self.task.description)
        self.assertIsInstance(self.task.start_date, datetime)
        self.assertIsInstance(self.task.due_date, datetime)
        self.assertEqual(activity_type, self.task.activity_type)
        self.assertEqual(classification, self.task.classification)

    def test_create_on_scope(self):
        self.task = self.project.create_activity("__Test task")

        self.assertIsInstance(self.task, Activity)
        self.assertEqual(ActivityType.TASK, self.task.activity_type)
        self.assertEqual(ActivityClassification.WORKFLOW, self.task.classification)

    def test_create_below_parent(self):
        self.process.children()  # populate `_cached_children`.
        self.assertIsNotNone(
            self.process._cached_children, "Cached children should be an (empty) list."
        )

        new_task = self.process.create(
            name="__Testing task",
            activity_type=ActivityType.TASK,
        )

        current_children = self.process.children()

        self.assertTrue(current_children)
        self.assertIn(
            new_task,
            current_children,
            msg="New child task should be among the children.",
        )

    # NOTE: Change this test once we release the FORMS backend to production or a testing setup.
    @skip(
        "This test does not work as the testing backend does not have FORMS in the codebase"
    )
    def test_create_with_classification(self):

        for classification in ActivityClassification.values():
            with self.subTest(msg=f"Classification: {classification}"):
                # setUp 1
                root_name = activity_root_name_by_classification[classification]
                root = self.project.activity(name=root_name)

                # testing 1
                self.assertEqual(classification, root.classification)
                self.assertEqual(ActivityType.PROCESS, root.activity_type)

                # setUp 2
                task = self.client.create_activity(
                    parent=root,
                    name=f"{classification}",
                    classification=classification,
                )

                # testing 2
                self.assertEqual(classification, task.classification)

                # tearDown
                task.delete()

    def test_create_with_incorrect_classification(self):
        with self.assertRaises(IllegalArgumentError):
            self.project.create_activity(
                name="Impossible classification",
                classification="Gummy bears",
            )

    def test_create_with_incorrect_parent(self):
        with self.assertRaises(IllegalArgumentError):
            self.client.create_activity(
                name="Impossible parent",
                parent="Darth vader",
            )

    def test_create_with_task_as_parent(self):
        task = self.process.create(name="__Test task")

        with self.assertRaises(
            IllegalArgumentError, msg="Tasks cannot be created below other tasks!"
        ):
            task.create("This cannot happen")

    def test_create_with_incorrect_inputs(self):
        with self.assertRaises(IllegalArgumentError):
            self.project.create_activity("__test_task", status="COMPLETE")

        with self.assertRaises(IllegalArgumentError):
            self.project.create_activity("__test_task", start_date=4)

        with self.assertRaises(IllegalArgumentError):
            self.project.create_activity("__test_task", description=1234)

        with self.assertRaises(IllegalArgumentError):
            self.project.create_activity("__test_task", classification="PRODUCT")

    def test_delete(self):
        # setUp
        sub_process_name = "__Test subprocess"
        sub_task_name = "__Test subtask"

        subprocess = self.process.create(
            name=sub_process_name, activity_type=ActivityType.PROCESS
        )
        self.task = subprocess.create(name=sub_task_name)
        subprocess.delete()

        # testing
        with self.assertRaises(APIError, msg="Cant delete the same Activity twice!"):
            subprocess.delete()
        with self.assertRaises(NotFoundError, msg="Deleted Activity cant be found!"):
            self.project.activity(name=sub_process_name)
        with self.assertRaises(
            NotFoundError, msg="Children of deleted Activities cant be found!"
        ):
            self.project.activity(name=sub_task_name)


class TestActivityClone(TestBetamax):
    def setUp(self):
        super().setUp()
        self.process = self.project.create_activity(
            name="__TEST CLONE SUBPROCESS", activity_type=ActivityType.PROCESS
        )
        self.task = self.process.create("__TEST CLONE TASK")
        self.clone = None
        self.bucket = [self.process]

    def tearDown(self):
        for activity in self.bucket:
            if activity:
                try:
                    activity.delete()
                except APIError:
                    pass
        super().tearDown()

    def test(self):
        clone = self.task.clone()

        self.assertIsInstance(clone, Activity)
        self.assertNotEqual(self.task, clone)
        self.assertEqual(self.task.parent_id, clone.parent_id)

    def test_parent_id(self):
        second_process = self.project.create_activity(
            name="__Test process 2",
            activity_type=ActivityType.PROCESS,
        )
        self.bucket.append(second_process)

        clone = self.task.clone(
            parent=second_process,
        )

        self.assertNotEqual(self.task.parent_id, clone.parent_id)

    def test_update(self):
        new_name = "__TEST TASK RENAMED"
        clone = self.task.clone(
            update_dict=dict(name=new_name),
        )

        self.assertEqual(new_name, clone.name)

    def test_update_incorrect(self):
        with self.assertRaises(IllegalArgumentError):
            self.task.clone(part_parent_instance=True)

    def test_async_via_task(self):
        response = self.task.clone(asynchronous=True)

        self.assertIsNone(response)

    def test_async_via_client(self):
        response = self.client.clone_activities(
            activities=[self.task], activity_parent=self.process, asynchronous=True
        )

        self.assertIsInstance(response, list)
        self.assertFalse(response)


class TestActivityCloneParts(TestBetamax):
    def setUp(self):
        super().setUp()

        # Create task to clone
        self.process = self.project.create_activity(
            name="__TEST CLONE SUBPROCESS", activity_type=ActivityType.PROCESS
        )
        self.task = self.process.create(name="__TEST CLONE TASK")

        # Create part model to copy along
        intermediate = self.project.catalog_root_model.add_model(
            name="__TEST CLONE INTERMEDIATE MODEL", multiplicity=Multiplicity.ONE
        )
        source_parent_model = intermediate.add_model(
            name="__TEST CLONE CONFIGURED MODEL - PARENT",
            multiplicity=Multiplicity.ONE_MANY,
        )
        child_model = source_parent_model.add_model(
            name="__TEST CLONE CONFIGURED MODEL - CHILD",
            multiplicity=Multiplicity.ONE_MANY,
        )

        for prop_type in [
            PropertyType.CHAR_VALUE,
            PropertyType.DATE_VALUE,
        ]:
            child_model.add_property(
                name="__TEST " + prop_type, property_type=prop_type
            )

        # Add widget to add configured part models
        wm = self.task.widgets()
        wm.add_filteredgrid_widget(
            parent_instance=source_parent_model.instance(),
            part_model=child_model,
            all_readable=True,
        )

        self.bike_model = self.project.model("Bike")
        self.bike_instance = self.bike_model.instance()
        wm.add_propertygrid_widget(
            part_instance=self.bike_instance,
            all_readable=True,
        )

        # Create target parents to move to
        self.target_parent_model = self.project.product_root_model.add_model(
            name="__TEST CLONE TARGET PARENT", multiplicity=Multiplicity.ONE
        )
        self.parent_instance = self.target_parent_model.instance()

        # In tearDown, delete tasks first, then configured data models
        self.bucket = [self.task, self.process, intermediate, self.target_parent_model]

    def tearDown(self):
        for obj in self.bucket:
            if obj:
                try:
                    obj.delete()
                except APIError:
                    pass
        super().tearDown()

    def test(self):
        """Copy a data model from the catalog to the product data model tree"""
        clones = self.client.clone_activities(
            activities=[self.task],
            activity_parent=self.process,
            activity_update_dicts={
                self.task.id: {"name": "__TEST CLONE ACTIVITY WITH PARTS"}
            },
            include_part_models=True,
            include_part_instances=True,
            include_children=True,
            part_parent_model=self.target_parent_model,
            part_parent_instance=self.parent_instance,
            asynchronous=False,
        )

        self.assertTrue(clones)

        new_children = list(self.parent_instance.children())
        self.assertTrue(new_children, msg="No parts were copied")

    def test_excluded_models(self):
        """Exclude the bike model from the copy"""
        clones = self.client.clone_activities(
            activities=[self.task],
            activity_parent=self.process,
            activity_update_dicts={
                self.task.id: {"name": "__TEST CLONE ACTIVITY WITH PARTS"}
            },
            include_part_models=True,
            include_part_instances=True,
            include_children=True,
            excluded_parts=[self.bike_model, self.bike_instance],
            part_parent_model=self.target_parent_model,
            part_parent_instance=self.parent_instance,
            asynchronous=False,
        )

        self.assertTrue(clones)

        new_children = list(self.parent_instance.children())
        self.assertTrue(new_children, msg="No parts were copied")
        self.assertNotIn(
            "Bike",
            {c.name for c in new_children},
            msg=(
                "Bike should not have been copied over. "
                "Actually it is not copied over, it is moved to the parent_instance"
            ),
        )


class TestActivities(TestBetamax):
    NAME = "__TEST ACTIVITY"

    def setUp(self):
        super().setUp()
        self.workflow_root = self.project.activity(name=ActivityRootNames.WORKFLOW_ROOT)
        self.task = self.project.create_activity(
            name=self.NAME, activity_type=ActivityType.TASK
        )

    def tearDown(self):
        if self.task:
            try:
                self.task.delete()
            except APIError:
                pass
        super().tearDown()

    def test_retrieve_activities(self):
        self.assertTrue(self.project.activities())

    def test_retrieve_single_activity(self):
        self.assertTrue(self.project.activity(self.NAME))

    def test_activity_attributes(self):
        attributes = [
            "_client",
            "_json_data",
            "id",
            "name",
            "created_at",
            "updated_at",
            "ref",
            "description",
            "status",
            "activity_type",
            "_scope_id",
            "start_date",
            "due_date",
        ]

        for attribute in attributes:
            with self.subTest(msg=attribute):
                self.assertTrue(
                    hasattr(self.task, attribute),
                    "Could not find '{}' in the object: '{}'".format(
                        attribute, self.task.__dict__.keys()
                    ),
                )

    def test_retrieve_unknown_activity(self):
        with self.assertRaises(NotFoundError):
            self.project.activity("Hello?!")

    def test_retrieve_too_many_activity(self):
        with self.assertRaises(MultipleFoundError):
            self.project.activity()

    # new in 1.7
    def test_edit_activity_name(self):
        self.task.edit(name="Specify wheel diameter - updated")

        self.task_u = self.project.activity("Specify wheel diameter - updated")
        self.assertEqual(self.task.id, self.task_u.id)
        self.assertEqual(self.task.name, self.task_u.name)
        self.assertEqual(self.task.name, "Specify wheel diameter - updated")

        # Added to improve coverage. Assert whether IllegalArgumentError is raised when 'name' is not a string object.
        with self.assertRaises(IllegalArgumentError):
            self.task.edit(name=True)

    def test_edit_activity_description(self):
        self.task.edit(description="This task has a cool description")

        self.assertEqual(self.task._client.last_response.status_code, requests.codes.ok)

        # Added to improve coverage. Assert whether IllegalArgumentError is raised when 'description' is
        # not a string object.
        with self.assertRaises(IllegalArgumentError):
            self.task.edit(description=42)

    def test_edit_activity_naive_dates(self):
        start_time = datetime(2000, 1, 1, 0, 0, 0)
        due_time = datetime(2019, 12, 31, 0, 0, 0)

        with warnings.catch_warnings(record=False):
            warnings.simplefilter("ignore")
            self.task.edit(start_date=start_time, due_date=due_time)

        self.assertEqual(self.task._client.last_response.status_code, requests.codes.ok)

        with self.assertRaises(IllegalArgumentError):
            self.task.edit(start_date="All you need is love")

        with self.assertRaises(IllegalArgumentError):
            self.task.edit(due_date="Love is all you need")

    def test_edit_due_date_timezone_aware(self):
        self.task.edit(start_date=self.time, due_date=self.time)

        self.assertEqual(self.task._client.last_response.status_code, requests.codes.ok)

    # 1.10.0
    def test_edit_activity_status(self):
        self.task.edit(status=ActivityStatus.COMPLETED)

        for status in [True, "NO STATUS", 3]:
            with self.subTest(msg=status):
                with self.assertRaises(IllegalArgumentError):
                    self.task.edit(status=status)

    # 1.7.2
    def test_datetime_with_naive_duedate_only_fails(self):
        """reference to #121 - thanks to @joost.schut"""
        naive_duedate = datetime(2017, 6, 5, 5, 0, 0)

        with warnings.catch_warnings(record=False):
            warnings.simplefilter("ignore")
            self.task.edit(due_date=naive_duedate)

    def test_datetime_with_tzinfo_provides_correct_offset(self):
        """reference to #121 - thanks to @joost.schut

        The tzinfo.timezone('Europe/Amsterdam') should provide a 2 hour offset, recording 20 minutes
        """
        # setup
        tz = pytz.timezone("Europe/Amsterdam")
        tzaware_due = tz.localize(datetime(2017, 7, 1))
        tzaware_start = tz.localize(datetime(2017, 6, 30, 0, 0, 0))

        self.task.edit(start_date=tzaware_start)
        self.assertTrue(
            self.task._json_data["start_date"], tzaware_start.isoformat(sep="T")
        )

        self.task.edit(due_date=tzaware_due)
        self.assertTrue(
            self.task._json_data["due_date"], tzaware_due.isoformat(sep="T")
        )

    def test_edit_cascade_down(self):
        # setup
        subprocess = self.project.activity("Subprocess")  # type: Activity
        subtask = self.project.activity("SubTask")  # type: Activity
        testuser = self.client.user(username="testuser")

        subprocess.edit_cascade_down(
            assignees=["testuser"],
            status=ActivityStatus.COMPLETED,
            overwrite=False,
        )

        subprocess.refresh()
        subtask.refresh()

        # testing
        self.assertIn(testuser, subprocess.assignees)
        self.assertIn(testuser, subtask.assignees)
        self.assertEqual(subprocess.status, ActivityStatus.COMPLETED)
        self.assertEqual(subtask.status, ActivityStatus.COMPLETED)

        # tearDown
        subprocess.edit(assignees=[], status=ActivityStatus.OPEN)
        subtask.edit(assignees=[], status=ActivityStatus.OPEN)

    # test added due to #847 - providing no inputs overwrites values
    def test_edit_activity_clearing_values(self):
        # setup
        initial_name = "Pykechain testing task"
        initial_description = "Task created to test editing."
        initial_start_date = datetime(2018, 12, 5, tzinfo=None)
        initial_due_date = datetime(2018, 12, 8, tzinfo=None)
        initial_tags = ["tag_one", "tag_two"]
        initial_assignee = self.client.user(username="testuser")

        self.task.edit(
            name=initial_name,
            description=initial_description,
            tags=initial_tags,
            start_date=initial_start_date,
            due_date=initial_due_date,
            assignees=[initial_assignee.username],
        )

        # Edit without mentioning values, everything should stay the same
        new_name = "New name for task"
        self.task.edit(name=new_name)

        # testing
        self.assertEqual(self.task.name, new_name)
        self.assertEqual(self.task.description, initial_description)
        self.assertEqual(
            self.task.start_date.strftime("%Y/%m/%d, %H:%M:%S"),
            initial_start_date.strftime("%Y/%m/%d, %H:%M:%S"),
        )
        self.assertEqual(
            self.task.due_date.strftime("%Y/%m/%d, %H:%M:%S"),
            initial_due_date.strftime("%Y/%m/%d, %H:%M:%S"),
        )
        self.assertEqual(self.task.tags, initial_tags)

        # Edit with clearing the values, name and status cannot be cleared
        self.task.edit(
            name=None,
            description=None,
            tags=None,
            start_date=None,
            due_date=None,
            status=None,
            assignees=None,
        )
        self.task.refresh()
        self.assertEqual(self.task.name, new_name)
        self.assertEqual(self.task.description, "")
        self.assertEqual(self.task.start_date, None)
        self.assertEqual(self.task.due_date, None)
        self.assertEqual(self.task.assignees, list())
        self.assertEqual(self.task.tags, list())

    def test_retrieve_children_of_task_fails_for_task(self):
        with self.assertRaises(NotFoundError, msg="Tasks have no children!"):
            self.task.children()

    def test_child(self):
        child_task = self.workflow_root.child(name=self.NAME)

        self.assertIsInstance(child_task, Activity)
        self.assertEqual(child_task._json_data["parent_id"], self.workflow_root.id)

    def test_child_invalid(self):
        with self.assertRaises(IllegalArgumentError):
            self.workflow_root.child()

        second_process = self.workflow_root.create(name=self.NAME)
        with self.assertRaises(MultipleFoundError):
            self.workflow_root.child(name=self.NAME)
        second_process.delete()

        with self.assertRaises(NotFoundError):
            self.workflow_root.child(name="Just a scratch")

    def test_retrieve_all_children(self):
        all_tasks = self.workflow_root.all_children()

        self.assertIsInstance(all_tasks, list)
        self.assertEqual(
            13, len(all_tasks), msg="Number of tasks has changed, expected 13."
        )

    def test_retrieve_activity_by_id(self):
        task = self.project.activity(name="Subprocess")  # type: Activity

        task_by_id = self.client.activity(pk=task.id)

        self.assertEqual(task.id, task_by_id.id)

    def test_retrieve_siblings_of_a_task_in_a_subprocess(self):
        task = self.project.activity(name="Subprocess")  # type: Activity
        siblings = task.siblings()

        self.assertIn(task.id, [sibling.id for sibling in siblings])
        self.assertTrue(len(siblings) >= 1)

    def test_retrieve_siblings_of_root(self):
        with self.assertRaises(NotFoundError):
            self.workflow_root.siblings()

    # in 1.12

    def test_retrieve_siblings_of_a_task_in_a_subprocess_with_arguments(self):
        task = self.project.activity(name="SubTask")  # type: Activity
        siblings = task.siblings(name__icontains="sub")

        self.assertIn(task.id, [sibling.id for sibling in siblings])
        self.assertEqual(1, len(siblings))

    def test_activity_without_scope_id_will_fix_itself(self):
        specify_wheel_diam_cripled = self.project.activity(
            name="Specify wheel diameter", fields="id,name,status"
        )
        self.assertFalse(specify_wheel_diam_cripled._json_data.get("scope_id"))

        # now the self-healing will begin
        self.assertEqual(specify_wheel_diam_cripled.scope_id, self.project.id)

    # in 1.13
    def test_create_activity_with_incorrect_activity_class_fails(self):
        with self.assertRaisesRegex(
            IllegalArgumentError, "must be an option from enum"
        ):
            self.project.create_activity(name="New", activity_type="DEFUNCTActivity")

    # 2.0 new activity
    # noinspection PyTypeChecker
    def test_edit_activity_assignee(self):
        specify_wd = self.project.activity("Specify wheel diameter")  # type: Activity
        original_assignee_ids = specify_wd._json_data.get("assignee_ids") or []

        # pykechain_user = self.client.user(username='pykechain')
        test_user = self.client.user(username="testuser")

        specify_wd.edit(assignees_ids=[test_user.id])
        specify_wd.refresh()

        self.assertIsInstance(specify_wd._json_data.get("assignees_ids")[0], int)

        self.assertEqual(
            specify_wd._client.last_response.status_code, requests.codes.ok
        )

        # Added to improve coverage. Assert whether NotFoundError is raised when 'assignee' is not part of the
        # scope members
        with self.assertRaises(NotFoundError):
            specify_wd.edit(assignees_ids=[-100])

        # Added to improve coverage. Assert whether NotFoundError is raised when 'assignee' is not part of the
        # scope members
        with self.assertRaises(IllegalArgumentError):
            specify_wd.edit(assignees_ids="this should have been a list")

        specify_wd.edit(assignees_ids=original_assignee_ids)

    def test_activity_retrieve_parent_of_task(self):
        task = self.project.activity(name="SubTask")
        subprocess = task.parent()  # type Activity
        self.assertEqual(subprocess.activity_type, ActivityType.PROCESS)

    def test_activity_retrieve_parent_of_root(self):
        task = self.project.activity(name=ActivityRootNames.WORKFLOW_ROOT)
        with self.assertRaises(NotFoundError):
            task.parent()

    def test_activity_retrieve_parent_of_a_toplevel_task_returns_workflow_root_id(
        self,
    ):
        task = self.project.activity("Specify wheel diameter")
        parent = task.parent()
        self.assertEqual(self.project._json_data.get("workflow_root_id"), parent.id)

    def test_activity_test_workflow_root_object(self):
        workflow_root = self.project.activity(
            id=self.project._json_data.get("workflow_root_id")
        )

        self.assertTrue(workflow_root.is_root())
        self.assertTrue(workflow_root.is_workflow_root())

    def test_activity_retrieve_children_of_parent(self):
        subprocess = self.project.activity(name="Subprocess")  # type: Activity
        children = subprocess.children()
        self.assertTrue(len(children) >= 1)
        for child in children:
            self.assertEqual(child._json_data.get("parent_id"), subprocess.id)

    def test_activity_retrieve_children_of_subprocess_with_arguments(self):
        subprocess = self.project.activity(name="Subprocess")  # type: Activity
        children = subprocess.children(name__icontains="task")
        self.assertTrue(len(children) >= 1)
        for child in children:
            self.assertEqual(child._json_data.get("parent_id"), subprocess.id)

    def test_count_children(self):
        process = self.project.activity(name="Tasks with Widgets")
        nr = process.count_children()

        self.assertIsInstance(nr, int)
        self.assertEqual(7, nr)

        nr = process.count_children(name__contains="Service")
        self.assertEqual(4, nr)

        with self.assertRaises(IllegalArgumentError):
            self.task.count_children()

    def test_rootlevel_activity_is_rootlevel(self):
        specify_wd = self.project.activity("Specify wheel diameter")

        self.assertTrue(specify_wd.is_rootlevel())

        root_itself = self.project.activity(ActivityRootNames.WORKFLOW_ROOT)

        self.assertFalse(root_itself.is_rootlevel())

    def test_subtask_activity_is_not_rootlevel(self):
        subprocess_subtask = self.project.activity("SubTask")

        self.assertFalse(subprocess_subtask.is_rootlevel())

    def test_activity_is_task(self):
        specify_wd = self.project.activity("Specify wheel diameter")

        self.assertTrue(specify_wd.is_task())
        self.assertFalse(specify_wd.is_subprocess())

    def test_activity_is_subprocess(self):
        subprocess = self.project.activity("Subprocess")

        self.assertTrue(subprocess.is_subprocess())
        self.assertFalse(subprocess.is_task())

    def test_activity_assignees_list(self):
        test_user = self.client.user(username="testuser")
        self.task.edit(assignees_ids=[test_user.id])
        self.task.refresh()

        list_of_assignees_in_data = self.task._json_data.get("assignees_ids")
        assignees_list = self.task.assignees

        self.assertSetEqual(
            set(list_of_assignees_in_data), {u.id for u in assignees_list}
        )

    def test_activity_assignees_list_no_assignees_gives_empty_list(self):
        activity_name = "Specify wheel diameter"
        activity = self.project.activity(name=activity_name)  # type: Activity

        self.assertListEqual(
            list(),
            activity.assignees,
            "Task has no assignees and should return Empty list",
        )

    def test_activity_move(self):
        # setUp
        activity_to_be_moved = self.task

        new_parent_name = "Subprocess"
        new_parent = self.project.activity(name=new_parent_name)

        activity_to_be_moved.move(parent=new_parent)

        # testing
        self.assertEqual(new_parent, activity_to_be_moved.parent())

    def test_activity_move_under_task_parent(self):
        # setUp
        new_parent_name = "Specify wheel diameter"
        new_parent = self.project.activity(name=new_parent_name)

        # testing
        with self.assertRaises(IllegalArgumentError):
            self.task.move(parent=new_parent)

    def test_activity_move_under_part_object(self):
        # setUp
        new_parent_name = "Bike"
        new_parent = self.project.part(name=new_parent_name)

        # testing
        with self.assertRaises(IllegalArgumentError):
            self.task.move(parent=new_parent)

    # tests added in 3.0
    def test_activity_retrieve_with_refs(self):
        # setup
        test_task_ref = slugify_ref(self.task.name)
        test_task_activity = self.project.activity(ref=test_task_ref)

        # testing
        self.assertIsInstance(test_task_activity, Activity)
        self.assertEqual(self.task, test_task_activity)

    def test_activity_associated_parts(self):
        # setUp
        activity_name = "Task - Form + Tables + Service"
        activity = self.project.activity(name=activity_name)
        associated_models, associated_instances = activity.associated_parts()

        # testing
        for model in associated_models:
            self.assertTrue(model.category == Category.MODEL)
            if model.name == "Bike":
                self.assertTrue(model.property(name="Gears").output)
                self.assertFalse(model.property(name="Total height").output)
                self.assertFalse(model.property(name="Picture").output)
                self.assertFalse(model.property(name="Description").output)
                self.assertTrue(model.property(name="Website").output)
                self.assertTrue(model.property(name="Sale?").output)

        for instance in associated_instances:
            self.assertTrue(instance.category == Category.INSTANCE)
        self.assertTrue(len(associated_models) == 3)
        self.assertTrue(len(associated_instances) == 4)

    def test_activity_associated_objects_ids(self):
        # setUp
        activity_name = "Task - Form + Tables + Service"
        activity = self.project.activity(name=activity_name)
        associated_object_ids = activity.associated_object_ids()

        # testing
        self.assertTrue(len(associated_object_ids) == 17)

    def test_activity_parts_of_specific_type(self):
        # setUp
        activity_name = "Task - Form + Tables + Service"
        bike_model = self.project.model(name="Bike")
        activity = self.project.activity(name=activity_name)
        associated_models = activity.parts(category=Category.MODEL)

        # testing
        for model in associated_models:
            self.assertTrue(model.category == Category.MODEL)
            if model == bike_model:
                self.assertTrue(model.property(name="Gears").output)
                self.assertFalse(model.property(name="Total height").output)
                self.assertFalse(model.property(name="Picture").output)
                self.assertFalse(model.property(name="Description").output)
                self.assertTrue(model.property(name="Website").output)
                self.assertTrue(model.property(name="Sale?").output)
        self.assertTrue(len(associated_models) == 3)


class TestActivityDownloadAsPDF(TestBetamax):
    def test_activity_download_as_pdf(self):
        # setUp
        activity_name = "Task - Form"
        activity = self.project.activity(name=activity_name)
        user = self.client.user(id=3)  # testmanager

        # testing
        with temp_chdir() as target_dir:
            pdf_file = activity.download_as_pdf(
                target_dir=target_dir, pdf_filename="pdf_file", user=user
            )
            self.assertTrue(os.path.exists(pdf_file))

            pdf_file_called_after_activity = activity.download_as_pdf(
                target_dir=target_dir,
            )
            self.assertTrue(os.path.exists(pdf_file_called_after_activity))

    @pytest.mark.skipif(
        "os.getenv('TRAVIS', False) or os.getenv('GITHUB_ACTIONS', False)",
        reason="Skipping tests when using Travis or Github Actions, as not Auth can be provided",
    )
    def test_activity_download_as_pdf_async(self):
        activity_name = "Task - Form"
        activity = self.project.activity(name=activity_name)

        # testing
        with temp_chdir() as target_dir:
            pdf_file = activity.download_as_pdf(
                target_dir=target_dir,
                pdf_filename="pdf_file",
                include_appendices=True,
                include_qr_code=True,
            )
            self.assertTrue(os.path.exists(pdf_file))

    def test_activity_share_link(self):
        # setUp
        test_user = self.client.user(username="testuser")

        activity_name = "Task - Form"
        message = "EXAMPLE_MESSAGE"
        subject = "EXAMPLE_SUBJECT"
        recipient_users = [test_user]

        activity = self.project.activity(name=activity_name)

        activity.share_link(
            subject=subject,
            message=message,
            recipient_users=recipient_users,
        )

        # testing
        notifications = self.client.notifications(
            subject=subject,
            message=message,
            event=NotificationEvent.SHARE_ACTIVITY_LINK,
        )
        self.assertEqual(self.client.last_response.status_code, requests.codes.ok)
        self.assertTrue(len(notifications), 1)

        # tearDown
        notifications[0].delete()

    def test_activity_share_pdf(self):
        # setUp
        test_user = self.client.user(username="testuser")
        activity_name = "Task - Form"
        message = "EXAMPLE_MESSAGE"
        subject = "EXAMPLE_SUBJECT"
        paper_size = PaperSize.A2
        paper_orientation = PaperOrientation.PORTRAIT
        recipient_users = [test_user]

        activity = self.project.activity(name=activity_name)

        activity.share_pdf(
            subject=subject,
            message=message,
            recipient_users=recipient_users,
            paper_size=paper_size,
            paper_orientation=paper_orientation,
            include_appendices=False,
            include_qr_code=True,
        )

        # testing
        notifications = self.client.notifications(
            subject=subject, message=message, event=NotificationEvent.SHARE_ACTIVITY_PDF
        )
        self.assertEqual(self.client.last_response.status_code, requests.codes.ok)
        self.assertTrue(len(notifications), 1)

        # tearDown
        notifications[0].delete()

    def test_activity_share_pdf_with_from_user(self):
        # setUp
        test_user = self.client.user(username="anotheruser")
        from_user = self.client.user(username="testuser")
        activity_name = "Task - Form"
        message = "EXAMPLE_MESSAGE"
        subject = "EXAMPLE_SUBJECT"
        paper_size = PaperSize.A2
        paper_orientation = PaperOrientation.PORTRAIT
        recipient_users = [test_user]

        activity = self.project.activity(name=activity_name)

        activity.share_pdf(
            from_user=from_user,
            subject=subject,
            message=message,
            recipient_users=recipient_users,
            paper_size=paper_size,
            paper_orientation=paper_orientation,
            include_appendices=False,
            include_qr_code=True,
        )

        # testing
        notifications = self.client.notifications(
            subject=subject, message=message, event=NotificationEvent.SHARE_ACTIVITY_PDF
        )
        self.assertEqual(self.client.last_response.status_code, requests.codes.ok)
        self.assertTrue(len(notifications), 1)

        # tearDown
        notifications[0].delete()

    def test_activity_share_link_with_from_user(self):
        # setUp
        test_user = self.client.user(username="anotheruser")
        from_user = self.client.user(username="testuser")

        activity_name = "Task - Form"
        message = "EXAMPLE_MESSAGE"
        subject = "EXAMPLE_SUBJECT"
        recipient_users = [test_user]

        activity = self.project.activity(name=activity_name)

        activity.share_link(
            from_user=from_user,
            subject=subject,
            message=message,
            recipient_users=recipient_users,
        )

        # testing
        notifications = self.client.notifications(
            subject=subject,
            message=message,
            event=NotificationEvent.SHARE_ACTIVITY_LINK,
        )
        self.assertEqual(self.client.last_response.status_code, requests.codes.ok)
        self.assertTrue(len(notifications), 1)

        # tearDown
        notifications[0].delete()



import datetime
import os

from pykechain.utils import temp_chdir

from pykechain.models.expiring_download import ExpiringDownload
from tests.classes import TestBetamax


class TestExpiringDownloads(TestBetamax):
    def setUp(self):
        super().setUp()
        self.test_assets_dir = os.path.dirname(
            os.path.dirname(os.path.abspath(__file__)).replace("\\", "/")
        )
        self.test_content_path = os.path.join(
            self.test_assets_dir,
            "tests",
            "files",
            "test_upload_content_to_expiring_download",
            "test_upload_content.pdf",
        )
        self.now = datetime.datetime.now()

        self.test_expiring_download = self.client.create_expiring_download(
            expires_in=56000,
            expires_at=datetime.datetime.now(),
        )

    def tearDown(self):
        self.test_expiring_download.delete()
        super().tearDown()

    def test_create_expiring_download_with_content(self):
        content_path = os.path.join(
            self.test_assets_dir,
            "tests",
            "files",
            "test_upload_content_to_expiring_download",
            "test_upload_content.pdf",
        )
        new_expiring_download = self.client.create_expiring_download(
            expires_at=datetime.datetime.now(), expires_in=42000, content_path=content_path
        )
        self.assertTrue(isinstance(new_expiring_download, ExpiringDownload))
        new_expiring_download.delete()

    def test_retrieve_expiring_downloads(self):
        expiring_downloads = self.client.expiring_downloads()
        self.assertTrue(expiring_downloads)

    def test_update_expiring_download(self):
        self.test_expiring_download.edit(expires_in=42000)
        self.assertEqual(self.test_expiring_download.expires_in, 42000)

    def test_upload_expiring_download(self):
        upload_path = os.path.join(
            self.test_assets_dir,
            "tests",
            "files",
            "test_upload_content_to_expiring_download",
            "test_upload_content.pdf",
        )
        self.test_expiring_download.upload(content_path=upload_path)
        self.assertIsNotNone(self.test_expiring_download.filename)
        self.assertEqual(self.test_expiring_download.filename, "test_upload_content.pdf")

    def test_upload_wrong_content_path(self):
        upload_path = os.path.join(
            self.test_assets_dir,
            "tests",
            "files",
            "test_upload_content_to_expiring_download",
            "test_upload_content.py",
        )
        with self.assertRaises(OSError):
            self.test_expiring_download.upload(content_path=upload_path)

    def test_save_expiring_download_content(self):
        upload_path = os.path.join(
            self.test_assets_dir,
            "tests",
            "files",
            "test_upload_content_to_expiring_download",
            "test_upload_content.pdf",
        )
        self.test_expiring_download.upload(content_path=upload_path)
        with temp_chdir() as target_dir:
            self.test_expiring_download.save_as(target_dir=target_dir)
            self.assertEqual(len(os.listdir(target_dir)), 1)



import pytest
from envparse import env

from pykechain import get_project
from pykechain.enums import KechainEnv
from pykechain.exceptions import ClientError
from tests.classes import TestBetamax
from tests.utils import TEST_TOKEN, TEST_URL, TEST_SCOPE_NAME, TEST_SCOPE_ID

PSEUDO_TOKEN = "aabbccddeeffgg0011223344556677889900"
PSEUDO_PASSWORD = "abc123!@#"
PSEUDO_SCOPE_ID = "eeb0937b-da50-4eb2-8d74-f36259cca96e"


@pytest.mark.skipif(
    "os.getenv('TRAVIS', False) or os.getenv('GITHUB_ACTIONS', False)",
    reason="Skipping tests when using Travis or Github Actions, as not Auth can be provided",
)
class TestGetProjectHelperNotForTravis(TestBetamax):
    def setUp(self):
        super().setUp()
        # unset all environment variables in the self.env protected env world
        for kec_env in KechainEnv.values():
            self.env.unset(kec_env)

    def test_get_project__not_for_travis(self):
        self.env.set(KechainEnv.KECHAIN_FORCE_ENV_USE, "false")
        with self.env:
            project = get_project(TEST_URL, token=TEST_TOKEN, scope=TEST_SCOPE_NAME)
            self.assertEqual(project.name, TEST_SCOPE_NAME)

    # 1.12.1
    def test_get_project__force_env_use_no_vars(self):
        """Test the get_project by using KECHAIN_FORCE_ENV_USE=True"""
        self.env.set("KECHAIN_FORCE_ENV_USE", "True")
        with self.env:
            self.assertTrue(env.bool(KechainEnv.KECHAIN_FORCE_ENV_USE))

            with self.assertRaisesRegex(ClientError, "should be provided as environment variable"):
                # KECHAIN_SCOPE or KECHAIN_SCOPE_ID should be provided as environment variable
                get_project()

    def test_get_project__force_env_use__only_url(self):
        self.env.update(dict(KECHAIN_URL=TEST_URL, KECHAIN_FORCE_ENV_USE="True"))
        with self.env:
            self.assertTrue(env.bool(KechainEnv.KECHAIN_FORCE_ENV_USE))
            with self.assertRaisesRegex(ClientError, "should be provided as environment variable"):
                # Error: KECHAIN_SCOPE or KECHAIN_SCOPE_ID should be provided as environment variable
                get_project()

    def test_get_project__force_env_use__url_and_token(self):
        self.env.update(
            dict(KECHAIN_URL=TEST_URL, KECHAIN_TOKEN=TEST_TOKEN, KECHAIN_FORCE_ENV_USE="True")
        )
        with self.env:
            self.assertTrue(env.bool(KechainEnv.KECHAIN_FORCE_ENV_USE))
            with self.assertRaisesRegex(ClientError, "should be provided as environment variable"):
                # Error: KECHAIN_SCOPE or KECHAIN_SCOPE_ID should be provided as environment variable
                get_project()

    def test_get_project__force_env_use__url_token_and_name(self):
        self.env.update(
            dict(
                KECHAIN_URL=TEST_URL,
                KECHAIN_TOKEN=TEST_TOKEN,
                KECHAIN_SCOPE_ID=TEST_SCOPE_ID,
                KECHAIN_FORCE_ENV_USE="True",
            )
        )
        with self.env:
            self.assertTrue(env.bool(KechainEnv.KECHAIN_FORCE_ENV_USE))
            self.assertEqual(get_project().name, TEST_SCOPE_NAME)

    def test_get_project__force_env_use__other_things_provided(self):
        self.env.update(
            dict(
                KECHAIN_URL=TEST_URL,
                KECHAIN_TOKEN=TEST_TOKEN,
                KECHAIN_SCOPE_ID=TEST_SCOPE_ID,
                KECHAIN_FORCE_ENV_USE="True",
            )
        )
        with self.env:
            self.assertTrue(env.bool(KechainEnv.KECHAIN_FORCE_ENV_USE))
            self.assertEqual(
                get_project(url="http://whatever", token=PSEUDO_TOKEN).name, TEST_SCOPE_NAME
            )

    def test_test_get_project_with_scope_id__not_for_travis(self):
        project = get_project(TEST_URL, token=TEST_TOKEN, scope_id=TEST_SCOPE_ID)
        self.assertEqual(project.name, TEST_SCOPE_NAME)

    def test_get_project_from_env__not_for_travis(self):
        # setup
        self.env.update(
            dict(KECHAIN_URL=TEST_URL, KECHAIN_TOKEN=TEST_TOKEN, KECHAIN_SCOPE=TEST_SCOPE_NAME)
        )

        with self.env:
            project = get_project()
            self.assertEqual(project.name, self.env["KECHAIN_SCOPE"])


class TestGetProjectHelper(TestBetamax):
    def test_project_raises_error__no_auth(self):
        with self.assertRaisesRegex(ClientError, "Error: insufficient arguments"):
            get_project(url=TEST_URL)

    def test_project_raises_error__token_and_no_scope(self):
        with self.assertRaisesRegex(ClientError, "Error: insufficient arguments"):
            get_project(url=TEST_URL, token=TEST_TOKEN)

    def test_project_raises_error__no_pass(self):
        with self.assertRaisesRegex(ClientError, "Error: insufficient arguments"):
            get_project(url=TEST_URL, username="auser", scope="Scope")

    def test_project_raises_error__auth_and_no_scope(self):
        with self.assertRaisesRegex(ClientError, "Error: insufficient arguments"):
            get_project(url=TEST_URL, username="auser", password=PSEUDO_PASSWORD)

    def test_project_raises_error__scope_id_and_no_pass(self):
        with self.assertRaisesRegex(ClientError, "Error: insufficient arguments"):
            get_project(url=TEST_URL, username="auser", scope_id=PSEUDO_SCOPE_ID)

    def test_project_raises_error__auth_and_no_url(self):
        with self.assertRaisesRegex(ClientError, "Error: insufficient arguments"):
            get_project(username="auser", password=PSEUDO_PASSWORD, scope_id=PSEUDO_SCOPE_ID)

    def test_project_raises_error__token_and_no_url(self):
        with self.assertRaisesRegex(ClientError, "Error: insufficient arguments"):
            get_project(token=PSEUDO_TOKEN, scope_id=PSEUDO_SCOPE_ID)



from unittest import TestCase

from jsonschema import ValidationError, validate

from pykechain.enums import PropertyType, PropertyVTypes
from pykechain.exceptions import IllegalArgumentError
from pykechain.models import AttachmentProperty, Property
from pykechain.models.validators import (
    BooleanFieldValidator,
    InvalidVisualEffect,
    NumericRangeValidator,
    PropertyValidator,
    ValidVisualEffect,
    ValidatorEffect,
    VisualEffect,
)
from pykechain.models.validators.validator_schemas import (
    effects_jsonschema_stub,
    options_json_schema,
    validator_jsonschema_stub,
)
from pykechain.models.validators.validators import (
    AlwaysAllowValidator,
    EmailValidator,
    EvenNumberValidator,
    FileExtensionValidator,
    FileSizeValidator,
    OddNumberValidator,
    RegexStringValidator,
    RequiredFieldValidator,
    SingleReferenceValidator,
)
from tests.classes import TestBetamax


class TestValidatorJSON(TestCase):
    def test_valid_numeric_range_validator_json(self):
        options = dict(
            validators=[
                dict(
                    vtype="numericRangeValidator",
                    config=dict(
                        minvalue=2,
                        maxvalue=10,
                        stepsize=2,
                        on_valid=[dict(effect="CssEffect", config=dict(applyCss="valid"))],
                        on_invalid=[
                            dict(effect="CssEffect", config=dict(applyCss="invalid")),
                            dict(
                                effect="ErrorText",
                                config=dict(text="Range should be between 2 and 10 with step 2."),
                            ),
                        ],
                    ),
                )
            ]
        )

        validate(options_json_schema, options)

    def test_valid_requiredfield_validator_json(self):
        options = {
            "validators": [
                {
                    "vtype": "requiredFieldValidator",
                    "config": {
                        "effects_when_valid": [],
                        "effects_when_invalid": [
                            {"effect": "ErrorText", "config": {"Text": "This field is required."}}
                        ],
                    },
                },
                {"vtype": "numericRangeValidator", "config": {"minValue": 0}},
            ]
        }

        validate(options_json_schema, options)

    def test_valid_booleanfield_validator_json(self):
        options = {
            "validators": [
                {
                    "vtype": "booleanFieldValidator",
                    "config": {
                        "is_valid": False,
                        "is_invalid": [True, None],
                        "effects_when_valid": [],
                        "effects_when_invalid": [
                            {"effect": "ErrorText", "text": "Value of the field should be False"}
                        ],
                    },
                }
            ]
        }

        validate(options_json_schema, options)

    def test_validator_invalid_vtype(self):
        validator_json = {"vtype": "invalid", "config": {}}
        with self.assertRaisesRegex(ValidationError, "'invalid' is not one of"):
            validate(validator_json, validator_jsonschema_stub)

    def test_validator_missing_vtype(self):
        validator_json = {"config": {}}
        with self.assertRaisesRegex(ValidationError, "'vtype' is a required property"):
            validate(validator_json, validator_jsonschema_stub)

    def test_validator_config_not_a_dict(self):
        v = {"vtype": PropertyVTypes.NUMERICRANGE, "config": []}
        with self.assertRaisesRegex(ValidationError, "is not of type 'object'"):
            validate(v, validator_jsonschema_stub)

    def test_validator_on_valid_not_a_list(self):
        v = {"vtype": PropertyVTypes.NUMERICRANGE, "config": {"on_valid": {}}}
        with self.assertRaisesRegex(ValidationError, "is not of type 'array'"):
            validate(v, validator_jsonschema_stub)

    def test_validator_on_invalid_not_a_list(self):
        v = {"vtype": PropertyVTypes.NUMERICRANGE, "config": {"on_invalid": {}}}
        with self.assertRaisesRegex(ValidationError, "is not of type 'array'"):
            validate(v, validator_jsonschema_stub)

    def test_validator_on_valid_list_with_obj(self):
        v = {
            "vtype": PropertyVTypes.NUMERICRANGE,
            "config": {"on_valid": [{"effect": "", "config": {}}]},
        }
        validate(v, validator_jsonschema_stub)

    def test_validator_on_invalid_list_with_obj(self):
        v = {
            "vtype": PropertyVTypes.NUMERICRANGE,
            "config": {"on_invalid": [{"effect": "", "config": {}}]},
        }
        validate(v, validator_jsonschema_stub)

    def test_validatoreffect_requires_effect_property(self):
        v = {}
        with self.assertRaisesRegex(ValidationError, "'effect' is a required property"):
            validate(v, effects_jsonschema_stub)

    def test_validatoreffect_requires_config_property(self):
        v = {"effect": ""}
        with self.assertRaisesRegex(ValidationError, "'config' is a required property"):
            validate(v, effects_jsonschema_stub)

    def test_validatoreffect_not_allows_additional_properties(self):
        v = {"effect": "", "config": {}, "additional_option": None}
        with self.assertRaisesRegex(
            ValidationError,
            r"Additional properties are not allowed \('additional_option' was unexpected\)",
        ):
            validate(v, effects_jsonschema_stub)


class TestPropertyValidatorClass(TestCase):
    def test_propertyvalidator_produces_valid_json(self):
        pv = PropertyValidator()
        pv.validate_json()


class TestValidatorEffects(TestCase):
    def test_validator_effect_produces_valid_json(self):
        ve = ValidatorEffect()
        ve.validate_json()

    def test_visual_effect_produces_valid_json(self):
        ve = VisualEffect()
        self.assertTrue("applyCss" in ve.as_json().get("config"))
        ve.validate_json()

    def test_valid_visualeffect_produces_valid_json(self):
        ve = ValidVisualEffect()
        self.assertTrue("applyCss" in ve.as_json().get("config"))
        ve.validate_json()

    def test_invalid_visualeffect_productes_valid_json(self):
        ve = InvalidVisualEffect()
        self.assertTrue("applyCss" in ve.as_json().get("config"))
        ve.validate_json()


class TestValidatorParsing(TestCase):
    def test_valid_numeric_range_validator_json(self):
        validator_json = dict(
            vtype="numericRangeValidator",
            config=dict(
                minvalue=2,
                maxvalue=10,
                stepsize=2,
                enforce_stepsize=False,
                on_valid=[dict(effect="visualEffect", config=dict(applyCss="valid"))],
                on_invalid=[
                    dict(effect="visualEffect", config=dict(applyCss="invalid")),
                    dict(
                        effect="errorTextEffect",
                        config=dict(text="Range should be between 2 and 10 with step 2."),
                    ),
                ],
            ),
        )

        validator = PropertyValidator.parse(validator_json)

        self.assertIsInstance(validator, NumericRangeValidator)
        self.assertTrue(validator.validate_json)
        pass


class TestValidatorDumping(TestCase):
    def test_valid_numeric_range_validator_dumped(self):
        validator_json = dict(
            vtype="numericRangeValidator",
            config=dict(
                minvalue=2,
                maxvalue=10,
                stepsize=2,
                enforce_stepsize=False,
                on_valid=[dict(effect="visualEffect", config=dict(applyCss="valid"))],
                on_invalid=[
                    dict(effect="visualEffect", config=dict(applyCss="invalid")),
                    dict(
                        effect="errorTextEffect",
                        config=dict(text="Range should be between 2 and 10 with step 2."),
                    ),
                ],
            ),
        )

        validator = PropertyValidator.parse(validator_json)
        self.assertIsInstance(validator, NumericRangeValidator)

        dumped_json = validator.as_json()
        self.assertIsNone(validate(dumped_json, validator_jsonschema_stub))
        self.assertIsNone(validator.validate_json())


class TestNumericRangeValidator(TestCase):
    def test_numeric_range_without_settings_validated_json(self):
        validator = NumericRangeValidator()

        as_json = validator.as_json()
        self.assertIsNone(validator.validate_json())
        self.assertIsInstance(as_json, dict)
        self.assertDictEqual(as_json, {"config": {}, "vtype": "numericRangeValidator"})

    def test_numeric_range_validates_with_lower_bound(self):
        validator = NumericRangeValidator(minvalue=0)
        self.assertTrue(validator.is_valid(1))
        validator = NumericRangeValidator(minvalue=5)
        self.assertTrue(validator.is_valid(6))

    def test_numeric_range_validates_with_max_bound(self):
        validator = NumericRangeValidator(maxvalue=0)
        self.assertTrue(validator.is_valid(-3))
        validator = NumericRangeValidator(maxvalue=10)
        self.assertTrue(validator.is_valid(1))

    def test_numeric_range_validates_with_min_max_bound(self):
        validator = NumericRangeValidator(maxvalue=10, minvalue=0)
        self.assertTrue(validator.is_valid(1))

    def test_numeric_range_invalidates_with_lower_bound(self):
        validator = NumericRangeValidator(minvalue=0)
        self.assertFalse(validator.is_valid(-1))
        validator = NumericRangeValidator(minvalue=5)
        self.assertFalse(validator.is_valid(2))

    def test_numeric_range_invalidates_with_max_bound(self):
        validator = NumericRangeValidator(maxvalue=0)
        self.assertFalse(validator.is_valid(1))
        validator = NumericRangeValidator(maxvalue=5)
        self.assertFalse(validator.is_valid(6))

    def test_numeric_range_invalidates_with_min_max_bound(self):
        validator = NumericRangeValidator(maxvalue=10, minvalue=0)
        self.assertFalse(validator.is_valid(-1))

    def test_numeric_range_validates_with_min_max_bound_float(self):
        validator = NumericRangeValidator(maxvalue=10.0, minvalue=0.0)
        self.assertTrue(validator.is_valid(1.5))

    def test_numeric_range_validates_with_stepsize_float(self):
        validator = NumericRangeValidator(stepsize=0.2, enforce_stepsize=True)
        self.assertTrue(validator.is_valid(1.2))
        self.assertTrue(validator.is_invalid(1.3))

    def test_numeric_range_validates_with_stepsize_int_with_minvalue(self):
        validator = NumericRangeValidator(minvalue=11, stepsize=2, enforce_stepsize=True)
        self.assertTrue(validator.is_valid(13))
        self.assertTrue(validator.is_invalid(16))

    def test_numeric_range_raises_exception_when_min_is_greater_than_max(self):
        with self.assertRaisesRegex(Exception, "should be smaller than the maxvalue"):
            NumericRangeValidator(minvalue=11, maxvalue=-11)

    def test_numeric_range_raises_exception_when_enforce_stepsize_without_stepsize(self):
        with self.assertRaisesRegex(
            Exception, "The stepsize should be provided when enforcing stepsize"
        ):
            NumericRangeValidator(stepsize=None, enforce_stepsize=True)

    def test_numeric_range_does_not_respect_max_issue_435(self):
        """With reference to issue #435

        In KE-chain I've added a property with numeric range validators min = unbound and max = 1000.
        A value of 3333 simply returns as valid in pykechain:
        _validation_reasons =>'Value \'3333\' is between -inf and inf'
        """
        validator = NumericRangeValidator(maxvalue=1000)
        self.assertFalse(validator.is_valid(3333))
        self.assertRegex(validator.get_reason(), "Value '3333' should be between -inf and 1000")


class TestBooleanFieldValidator(TestCase):
    def test_boolean_validator_without_settings(self):
        validator = BooleanFieldValidator()
        self.assertIsNone(validator.validate_json())
        self.assertIsInstance(validator.as_json(), dict)
        self.assertDictEqual(validator.as_json(), {"config": {}, "vtype": "booleanFieldValidator"})


class TestRequiredFieldValidator(TestCase):
    def test_requiredfield_validator_without_settings(self):
        validator = RequiredFieldValidator()
        self.assertIsNone(validator.validate_json())
        self.assertIsInstance(validator.as_json(), dict)
        self.assertDictEqual(
            validator.as_json(), {"config": {}, "vtype": "requiredFieldValidator"}
        )

    def test_requiredfield_validator_is_false_on_nonevalue(self):
        validator = RequiredFieldValidator()
        self.assertFalse(validator.is_valid(None))
        self.assertFalse(validator.is_valid(""))
        self.assertFalse(validator.is_valid([]))
        self.assertFalse(validator.is_valid(list()))
        self.assertFalse(validator.is_valid(tuple()))
        self.assertFalse(validator.is_valid(set()))

        # the value 'False' is not a None value so that is True (placed in other test)
        # self.assertFalse(validator.is_valid(False))

    def test_requiredfield_validator_is_true_on_value(self):
        validator = RequiredFieldValidator()
        self.assertTrue(validator.is_valid(1))
        self.assertTrue(validator.is_valid(0))
        self.assertTrue(validator.is_valid(0.0))
        self.assertTrue(validator.is_valid(float("inf")))
        self.assertTrue(validator.is_valid("string"))
        self.assertTrue(validator.is_valid({"key": "val"}))
        self.assertTrue(validator.is_valid([1, 2, 3]))
        self.assertTrue(validator.is_valid((1,)))
        self.assertTrue(validator.is_valid({1, 2, 3}))
        self.assertTrue(validator.is_valid(False))


class TestRegexValidator(TestCase):
    def test_regex_validator_without_settings(self):
        validator = RegexStringValidator()
        self.assertIsNone(validator.validate_json())
        self.assertIsInstance(validator.as_json(), dict)
        self.assertDictEqual(validator.as_json(), {"config": {}, "vtype": "regexStringValidator"})

    def test_regex_validator_with_pattern_match(self):
        validator = RegexStringValidator(pattern=r".*")

        self.assertTrue(validator("mr cactus is tevree"))

    def test_regex_validator_without_pattern_match(self):
        validator = RegexStringValidator()

        # per default the regex string matches everything
        self.assertEqual(validator.pattern, ".+")
        self.assertTrue(validator("mr cactus is tevree"))
        self.assertFalse(validator(""))

    def test_regex_validator_fails_on_none_value(self):
        validator = RegexStringValidator(pattern=r".*")

        self.assertFalse(validator.is_valid(None))

    def test_regex_validator_complex_email_regex(self):
        validator = EmailValidator()

        self.assertTrue(validator.is_valid("support@ke-works.com"))
        self.assertFalse(validator.is_valid("___"))
        self.assertIsNone(validator.is_valid(None))
        self.assertFalse(validator.is_valid("user@domain"))


class TestOddEvenNumberValidator(TestCase):
    def test_even_number_validator_without_settings(self):
        validator = EvenNumberValidator()
        self.assertIsNone(validator.validate_json())
        self.assertIsInstance(validator.as_json(), dict)
        self.assertDictEqual(validator.as_json(), {"config": {}, "vtype": "evenNumberValidator"})

    def test_odd_number_validator_without_settings(self):
        validator = OddNumberValidator()
        self.assertIsNone(validator.validate_json())
        self.assertIsInstance(validator.as_json(), dict)
        self.assertDictEqual(validator.as_json(), {"config": {}, "vtype": "oddNumberValidator"})

    def test_even_number_validator_is_valid(self):
        validator = EvenNumberValidator()
        self.assertTrue(validator.is_valid(2))

    def test_odd_number_validator_is_valid(self):
        validator = OddNumberValidator()
        self.assertTrue(validator.is_valid(3))

    def test_even_number_validator_is_invalid(self):
        validator = EvenNumberValidator()
        self.assertFalse(validator.is_valid(3))

    def test_odd_number_validator_is_invalid(self):
        validator = OddNumberValidator()
        self.assertFalse(validator.is_valid(4))

    def test_even_number_validator_is_none(self):
        validator = EvenNumberValidator()
        self.assertIsNone(validator.is_valid(None))

    def test_odd_number_validator_is_none(self):
        validator = OddNumberValidator()
        self.assertIsNone(validator.is_valid(None))

    def test_even_number_validator_float_valid(self):
        validator = EvenNumberValidator()
        self.assertTrue(validator.is_valid(4.5))
        self.assertTrue(validator.is_valid(4.0))
        self.assertTrue(validator.is_valid(3.9999999999999999))  # rounding accuracy

    def test_odd_number_validator_float_valid(self):
        validator = OddNumberValidator()
        self.assertTrue(validator.is_valid(3.0))
        self.assertTrue(validator.is_valid(4.9999999999999999))  # rounding accuracy

    def test_even_number_validator_float_invalid(self):
        validator = EvenNumberValidator()
        self.assertFalse(validator.is_valid(3.99))
        self.assertFalse(validator.is_valid(5.5))

    def test_odd_number_validator_float_invalid(self):
        validator = OddNumberValidator()
        self.assertFalse(validator.is_valid(4.99))
        self.assertFalse(validator.is_valid(6.0))

    def test_even_number_validator_invalid_input(self):
        validator = EvenNumberValidator()
        self.assertFalse(validator.is_valid(dict()))
        self.assertFalse(validator.is_valid(list))
        self.assertFalse(validator.is_valid(set()))
        self.assertFalse(validator.is_valid(tuple()))
        self.assertFalse(validator.is_valid("3"))

    def test_odd_number_validator_invalid_input(self):
        validator = OddNumberValidator()
        self.assertFalse(validator.is_valid(dict()))
        self.assertFalse(validator.is_valid(list))
        self.assertFalse(validator.is_valid(set()))
        self.assertFalse(validator.is_valid(tuple()))
        self.assertFalse(validator.is_valid("3"))


class TestSingleReferenceValidator(TestCase):
    def test_singlereference_validator_without_settings(self):
        validator = SingleReferenceValidator()
        self.assertIsNone(validator.validate_json())
        self.assertIsInstance(validator.as_json(), dict)
        self.assertDictEqual(
            validator.as_json(), {"config": {}, "vtype": "singleReferenceValidator"}
        )

    def test_singlereference_validator_is_valid(self):
        validator = SingleReferenceValidator()
        self.assertIsNone(validator.is_valid(None))
        self.assertTrue(validator.is_valid(list()))
        self.assertTrue(validator.is_valid(set()))
        self.assertTrue(validator.is_valid(tuple()))
        self.assertTrue(validator.is_valid(("first selection",)))

    def test_singlereference_validator_is_invalid(self):
        validator = SingleReferenceValidator()
        self.assertFalse(validator.is_valid(["first", "second"]))
        self.assertFalse(validator.is_valid((1, 2)))

    def test_singlerefence_validator_is_invalid_with_invalid_values(self):
        validator = SingleReferenceValidator()
        self.assertFalse(validator.is_valid("a string"))
        self.assertFalse(validator.is_valid(1.0))
        self.assertFalse(validator.is_valid(dict()))


class TestAlwaysAllowValidator(TestCase):
    def test_always_allow_validator_without_settings(self):
        validator = AlwaysAllowValidator()
        self.assertTrue(validator.is_valid("some text"))
        self.assertTrue(validator.is_valid(42))
        self.assertTrue(validator.is_valid(42.0))
        self.assertTrue(validator.is_valid(["a", "b"]))
        self.assertTrue(validator.is_valid(True))
        self.assertTrue(validator.is_valid(False))
        self.assertTrue(validator.is_valid(None))
        self.assertTrue(validator.is_valid({"a": "b"}))
        self.assertFalse(validator.is_invalid(None))
        self.assertTrue(validator.is_valid(list()))
        self.assertTrue(validator.is_valid(set()))
        self.assertTrue(validator.is_valid(tuple()))
        self.assertIsNone(validator.validate_json())


class TestFileSizeValidator(TestCase):
    def test_validator_valid_json_with_settings(self):
        validator = FileSizeValidator(max_size=100)
        self.assertIsNone(validator.validate_json())

    def test_validator_valid_json_without_settings(self):
        validator = FileSizeValidator()
        self.assertIsNone(validator.validate_json())

    def test_validator_valid_json_with_additional_arguments(self):
        validator = FileSizeValidator(max_size=100, another_argument=False)
        # another_argument is not found in the json
        validator.validate_json()
        validator.as_json()

    # noinspection PyTypeChecker
    def test_validator_invalid_arguments(self):
        with self.assertRaisesRegex(ValueError, "should be a number"):
            FileSizeValidator(max_size=set())
        with self.assertRaisesRegex(ValueError, "should be a number"):
            FileSizeValidator(max_size=list())
        with self.assertRaisesRegex(ValueError, "should be a number"):
            FileSizeValidator(max_size=dict())
        with self.assertRaisesRegex(ValueError, "should be a number"):
            FileSizeValidator(max_size=tuple())
        with self.assertRaisesRegex(ValueError, "should be a number"):
            FileSizeValidator(max_size=tuple())
        with self.assertRaisesRegex(ValueError, "should be a number"):
            FileSizeValidator(max_size="100")

    def test_filesizevalidator_being_valid(self):
        validator = FileSizeValidator(max_size=100)

        self.assertTrue(validator.is_valid(100))
        self.assertTrue(validator.is_valid(100.0))
        self.assertTrue(validator.is_valid(99))
        self.assertTrue(validator.is_valid(0))

        self.assertTrue(validator.is_valid("text"))
        self.assertTrue(validator.is_valid(list()))
        self.assertTrue(validator.is_valid(tuple()))
        self.assertTrue(validator.is_valid(dict()))
        self.assertTrue(validator.is_valid(set()))

    def test_filesizevalidator_being_invalid(self):
        validator = FileSizeValidator(max_size=100)

        self.assertFalse(validator.is_valid(101))
        self.assertFalse(validator.is_valid(101.0))
        self.assertFalse(validator.is_valid(-1))

    def test_filesizevalidator_being_none(self):
        validator = FileSizeValidator(max_size=100)

        self.assertIsNone(validator.is_valid(None))


class TestFileExtensionValidator(TestCase):
    def test_validator_valid_json_with_settings(self):
        validator = FileExtensionValidator(accept=[".csv"])
        self.assertIsNone(validator.validate_json())

    def test_validator_valid_json_without_settings(self):
        validator = FileExtensionValidator()
        self.assertIsNone(validator.validate_json())

    def test_fileextensionvalidator_on_extension(self):
        validator = FileExtensionValidator(accept=[".csv"])
        self.assertTrue(validator.is_valid("file.csv"))
        self.assertFalse(validator.is_valid("file.tsv"))
        self.assertFalse(validator.is_valid("file.txt"))

    def test_fileextensionvalidator_on_extensions(self):
        validator = FileExtensionValidator(accept=[".csv", ".pdf"])
        self.assertTrue(validator.is_valid("file.csv"))
        self.assertTrue(validator.is_valid("file.pdf"))
        self.assertFalse(validator.is_valid("file.docx"))

        with self.assertRaises(ValueError):
            # noinspection PyTypeChecker
            FileExtensionValidator(accept=True)

    def test_fileextensionvalidator_on_extensions_csv(self):
        validator = FileExtensionValidator(accept=".csv,.pdf")
        self.assertTrue(validator.is_valid("file.csv"))
        self.assertTrue(validator.is_valid("file.pdf"))
        self.assertFalse(validator.is_valid("file.docx"))

    def test_fileextensionvalidator_on_mimetype(self):
        validator = FileExtensionValidator(accept=["application/pdf"])
        self.assertTrue(validator.is_valid("file.pdf"))
        self.assertFalse(validator.is_valid("file.docx"))

    def test_fileextensionvalidator_on_expanded_mimetypes(self):
        validator = FileExtensionValidator(accept=["video/*"])
        self.assertTrue(validator.is_valid("file.mp4"))
        self.assertTrue(validator.is_valid("file.mov"))
        self.assertTrue(validator.is_valid("file.mpeg"))
        self.assertTrue(validator.is_valid("file.flv"))
        self.assertTrue(validator.is_valid("file.wmv"))
        self.assertTrue(validator.is_valid("file.webm"))

    def test_fileextensionvalidator_on_mixed_mode(self):
        validator = FileExtensionValidator(accept=[".csv", "application/pdf"])
        self.assertTrue(validator.is_valid("file.csv"))
        self.assertTrue(validator.is_valid("file.pdf"))
        self.assertFalse(validator.is_valid("file.docx"))

    def test_fileextensionvalidator_on_excel(self):
        validator = FileExtensionValidator(accept=[".xls", ".xlsx"])
        self.assertTrue(validator.is_valid("file.xls"))
        self.assertTrue(validator.is_valid("file.xlsx"))
        self.assertFalse(validator.is_valid("file.docx"))

    def test_fileextensionvalidator_being_none(self):
        validator = FileExtensionValidator(accept=["video/*"])

        self.assertIsNone(validator.is_valid(None))


class TestPropertyWithValidator(TestCase):
    def test_property_without_validator(self):
        prop = Property(json={}, client=None)
        self.assertIsNone(prop.is_valid)
        self.assertIsNone(prop.is_invalid)
        self.assertEqual(prop._validators, list())
        self.assertEqual(prop.validate(), list())

    def test_property_with_numeric_range_validator(self):
        prop_json = dict(
            value=1,
            value_options=dict(
                validators=[NumericRangeValidator(minvalue=0, maxvalue=10).as_json()]
            ),
        )
        prop = Property(json=prop_json, client=None)
        self.assertTrue(prop.is_valid)
        self.assertFalse(prop.is_invalid)
        self.assertTrue(prop.validate())

    def test_property_with_numeric_range_validator_value_is_none(self):
        prop_json = dict(
            value=None,
            value_options=dict(
                validators=[NumericRangeValidator(minvalue=0, maxvalue=10).as_json()]
            ),
        )
        prop = Property(json=prop_json, client=None)
        self.assertIsNone(prop.is_valid)
        self.assertIsNone(prop.is_invalid)
        self.assertListEqual([(None, "No reason")], prop.validate())

    def test_property_with_filesize_validator(self):
        prop_json = dict(
            value="attachments/12345678-1234-5678-1234-567812345678/some_file.txt",
            value_options=dict(validators=[FileSizeValidator(max_size=100).as_json()]),
        )
        prop = AttachmentProperty(json=prop_json, client=None)
        self.assertTrue(prop.is_valid)
        self.assertFalse(prop.is_invalid)
        self.assertTrue(prop.validate())

    def test_property_without_value_with_filesize_validator(self):
        prop_json = dict(
            value=None, value_options=dict(validators=[FileSizeValidator(max_size=100).as_json()])
        )
        prop = AttachmentProperty(json=prop_json, client=None)
        self.assertIsNone(prop.is_valid)
        self.assertIsNone(prop.is_invalid)
        self.assertListEqual([(None, "No reason")], prop.validate())

    def test_property_with_fileextension_validator(self):
        prop_json = dict(
            value="attachments/12345678-1234-5678-1234-567812345678/some_file.txt",
            value_options=dict(
                validators=[FileExtensionValidator(accept=[".txt", "application/pdf"]).as_json()]
            ),
        )
        prop = AttachmentProperty(json=prop_json, client=None)
        self.assertTrue(prop.is_valid)
        self.assertFalse(prop.is_invalid)
        self.assertTrue(prop.validate())

    def test_property_without_value_with_fileextension_validator(self):
        prop_json = dict(
            value=None,
            value_options=dict(
                validators=[FileExtensionValidator(accept=[".txt", "application/pdf"]).as_json()]
            ),
        )
        prop = AttachmentProperty(json=prop_json, client=None)
        self.assertIsNone(prop.is_valid)
        self.assertIsNone(prop.is_invalid)
        self.assertListEqual([(None, "No reason")], prop.validate())


class TestPropertyWithValidatorFromLiveServer(TestBetamax):
    def setUp(self):
        super().setUp()
        self.part_model = self.project.model(name="Model")
        self.numeric_range_prop_model = self.part_model.add_property(
            name="numericrange_validatortest", property_type=PropertyType.FLOAT_VALUE
        )
        self.numeric_range_prop_model.validators = [NumericRangeValidator(min=0.0, max=42.0)]
        self.part = self.project.part(name__startswith="Catalog").add(
            self.part_model, name="Model Instance for tests"
        )
        self.numeric_range_prop_instance = self.part.property(name="numericrange_validatortest")

    def tearDown(self):
        self.numeric_range_prop_model.delete()
        self.part.delete()
        super().tearDown()

    def test_numeric_property_with_validator_parses(self):
        self.assertIsInstance(self.numeric_range_prop_instance._validators, list)
        self.assertIsInstance(self.numeric_range_prop_instance._validators[0], PropertyValidator)

    def test_numeric_property_add_requiredvalidator_on_model(self):
        # test
        validators = self.numeric_range_prop_model.validators
        validators.append(RequiredFieldValidator())
        self.numeric_range_prop_model.validators = validators

        for validator in self.numeric_range_prop_model.validators:
            self.assertIsInstance(validator, (NumericRangeValidator, RequiredFieldValidator))

    def test_numeric_property_add_requiredvalidator_on_instance(self):
        # test
        validators = self.numeric_range_prop_instance.validators
        validators.append(RequiredFieldValidator())

        with self.assertRaisesRegex(
            IllegalArgumentError,
            "To update the list of validators, it can only work on `Property` of category 'MODEL'",
        ):
            self.numeric_range_prop_instance.validators = validators



from envparse import Env

# reads a local .env file with the TEST_TOKEN=<user token>
# ensure that this file is not commited to github (never ever)
env = Env()
env.read_envfile()

TEST_URL = env("TEST_URL", default="https://pim3-test.ke-chain.com")
TEST_USERNAME = env("TEST_USERNAME", default="pykechain")  # LVL1
TEST_TOKEN = env("TEST_TOKEN", default="")
TEST_SCOPE_ID = env("TEST_SCOPE_ID", default="bd5dceaa-a35e-47b0-9fc5-875410f4a56f")
TEST_SCOPE_NAME = env("TEST_SCOPE_NAME", default="Bike Project")
TEST_RECORD_CASSETTES = env.bool("TEST_RECORD_CASSETTES", default=True)



from pykechain.exceptions import _DeprecationMixin
from pykechain.models import SelectListProperty, MultiSelectListProperty


class SelectListProperty2(SelectListProperty, _DeprecationMixin):
    """A virtual object representing a KE-chain single-select list property."""

    pass


class MultiSelectListProperty2(MultiSelectListProperty, _DeprecationMixin):
    """A virtual object representing a KE-chain multi-select list property."""

    pass



from typing import Dict, Optional

from pykechain.enums import URITarget, FontAwesomeMode
from pykechain.exceptions import IllegalArgumentError

allowed_attributes = ["displayName_nl", "displayName_de", "displayName_fr", "displayName_it"]


class SideBarButton:
    """
    Side-bar button class.

    Every custom button in the side-bar is maintained as an object of this class.
    The original KE-chain buttons for the project detail, tasks and work breakdown structure are not separate buttons.
    """

    def __init__(
        self,
        side_bar_manager: "SideBarManager",
        json: Optional[Dict] = None,
        order: Optional[int] = None,
        title: Optional[str] = None,
        icon: Optional[str] = None,
        uri: Optional[str] = None,
        uri_target: URITarget = URITarget.INTERNAL,
        icon_mode: FontAwesomeMode = FontAwesomeMode.REGULAR,
        **kwargs,
    ):
        """
        Create a side-bar button.

        :param side_bar_manager: Manager object to which the button is linked.
        :type side_bar_manager: SideBarManager
        :param json: the json response to construct the :class:`SideBarButton` from
        :type json: dict
        :param order: index of the button
        :type order: int
        :param title: visible label of the button
        :type title: str
        :param icon: FontAwesome icon of the button
        :type icon: str
        :param uri: Uniform Resource Identifier, the address of the linked page
        :type uri: str
        :param uri_target: type of URI, either internal or external
        :type uri_target: URITarget
        :param icon_mode: FontAwesome display mode of the icon
        :type icon_mode: FontAwesomeMode
        :returns None
        :raises IllegalArgumentError: When the provided Argument is not according to the type.
        """
        super().__init__()

        if json is None:
            json = {}

        title = title if title else json.get("displayName")
        order = order if order is not None else json.get("order")
        icon = icon if icon else json.get("displayIcon")
        uri = uri if uri else json.get("uri")
        uri_target = json.get("uriTarget", uri_target)
        icon_mode = json.get("displayIconMode", icon_mode)

        if not isinstance(order, int):
            raise IllegalArgumentError(f'order must be an integer, "{order}" is not.')
        if not isinstance(title, str):
            raise IllegalArgumentError(f'title must be a string, "{title}" is not.')
        if not isinstance(icon, str):
            raise IllegalArgumentError(f'icon must be a string, "{icon}" is not.')
        if not isinstance(uri, str):
            raise IllegalArgumentError(f'uri must be a string, "{uri}" is not.')
        if uri_target not in URITarget.values():
            raise IllegalArgumentError(
                f'uri_target must be a URITarget option, "{uri_target}" is not.'
            )
        if icon_mode not in FontAwesomeMode.values():
            raise IllegalArgumentError(
                f'icon_mode must be a FontAwesomeMode option, "{icon_mode}" is not.'
            )

        for key in kwargs.keys():
            if key not in allowed_attributes:
                raise IllegalArgumentError(
                    f'Attribute "{key}" is not supported in the configuration of a side-bar'
                    " button."
                )

        self._manager: "SideBarManager" = side_bar_manager
        self.display_name: str = title
        self.order: int = order
        self.display_icon: str = icon
        self.uri: str = uri
        self.uri_target: URITarget = uri_target
        self.display_icon_mode: FontAwesomeMode = icon_mode

        self._other_attributes = kwargs
        for key in allowed_attributes:
            if key in json:
                self._other_attributes[key] = json[key]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} {self.order}: {self.display_name}"

    def refresh(self, json: Optional[Dict] = None) -> None:
        """
        Refresh the object in-place using the provided json.

        :param json: the json response to construct the :class:`SideBarButton` from
        :type json: dict
        :return: None
        """
        self.__init__(side_bar_manager=self._manager, json=json or self.as_dict())

    def edit(self, **kwargs) -> None:
        """
        Edit the details of the button.

        :param kwargs:
        :return: None
        """
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            elif key in allowed_attributes:
                self._other_attributes[key] = value

        self._manager._update()

    def delete(self) -> None:
        """
        Delete the side-bar button from the side-bar.

        :return: None
        """
        self._manager.delete_button(key=self)

    def as_dict(self) -> Dict:
        """
        Retrieve the configuration data, or `meta`, of the side-bar button.

        :return: dictionary of the configuration data
        :rtype dict
        """
        config = {
            "displayName": self.display_name,
            "displayIcon": self.display_icon,
            "uriTarget": self.uri_target,
            "uri": self.uri,
            "order": self.order,
            "displayIconMode": self.display_icon_mode,
        }
        config.update(self._other_attributes)

        return config



"""A python library to connect and interact with KE-chain."""
import sys

from .__about__ import version
from .client import Client
from .helpers import get_project

__all__ = ("Client", "get_project", "version")

if sys.version_info.major == 2 or (sys.version_info.major == 3 and sys.version_info.minor < 6):
    raise RuntimeError(
        "Python version >= `3.6` is required for this version of `pykechain` to operate. "
        "Please use `pykechain` version `2.7.x` for usage in combination with Python `2.7`"
    )



from collections.abc import Iterable
from typing import Any, Optional, List, Dict

from pykechain.enums import (
    URITarget,
    SubprocessDisplayMode,
    KEChainPages,
    KEChainPageLabels,
    KEChainPageIcons,
)
from pykechain.exceptions import NotFoundError
from pykechain.models.input_checks import (
    check_url,
    check_text,
    check_enum,
    check_list_of_dicts,
    check_type,
)
from pykechain.models.sidebar.sidebar_button import SideBarButton
from pykechain.utils import find


class SideBarManager(Iterable):
    """
    Sidebar manager class.

    :ivar scope: Scope object for which the side-bar is managed.
    :type scope: Scope
    :ivar bulk_creation: boolean to create buttons in bulk, postponing updating of KE-chain until the manager is
                         deleted from memory (end of your function)
    :type bulk_creation: bool
    """

    __existing_managers = (
        dict()
    )  # storage of manager objects to enforce 1 manager object per Scope

    def __new__(cls, scope: "Scope", *args, **kwargs):
        """Overwrite superclass method to enforce singleton manager per Scope object."""
        instance = super().__new__(cls)

        # Singleton manager per scope: this is required to support bulk_creation
        if scope.id in cls.__existing_managers:
            instance = cls.__existing_managers[scope.id]
        else:
            cls.__existing_managers[scope.id] = instance

        return instance

    def __init__(self, scope: "Scope", **kwargs):
        """
        Create a side-bar manager object for the Scope object.

        :param scope: Scope for which to create the side-bar manager.
        :type scope: Scope
        :param bulk_creation: flag whether to update once (True) or continuously (False, default)
        :type bulk_creation: bool
        """
        super().__init__(**kwargs)

        from pykechain.models import Scope

        check_type(scope, Scope, "scope")

        self.scope: Scope = scope
        self._override: bool = scope.options.get("overrideSideBar", False)

        self._scope_uri = f"#/scopes/{self.scope.id}"
        self._perform_bulk_creation = False

        self._buttons: List[SideBarButton] = []

        # Load existing buttons from the scope
        for button_dict in scope.options.get("customNavigation", []):
            self._buttons.append(SideBarButton(side_bar_manager=self, json=button_dict))

        self._iter = iter(self._buttons)

    def __repr__(self) -> str:  # pragma: no cover
        return f"<pyke {self.__class__.__name__} object {self.__len__()} buttons>"

    def __iter__(self):
        return self

    def __len__(self) -> int:
        return len(self._buttons)

    def __next__(self) -> SideBarButton:
        # py3.4 and up style next
        return next(self._iter)

    def __getitem__(self, key: Any) -> SideBarButton:

        found = None
        if isinstance(key, SideBarButton):
            found = find(self._buttons, lambda b: b == key)

        if isinstance(key, int):
            found = self._buttons[key]
        elif isinstance(key, str):
            found = find(self._buttons, lambda p: key == p.display_name)

        if found is not None:
            return found
        raise NotFoundError(f"Could not find button with index or name '{key}'")

    def __enter__(self):
        """
        Open context manager using the `with` keyword to postpone updates to KE-chain.

        >>> with scope.side_bar() as manager:
        >>>     button = manager.add_ke_chain_page(page_name=KEChainPages.EXPLORER)
        >>>     manager.insert(index=0, button=button)
        """
        self._perform_bulk_creation = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._perform_bulk_creation = False
        self._update()

    def refresh(self) -> None:
        """Reload the scope options from KE-chain to refresh the side-bar data."""
        self.scope.refresh()
        self.__init__(scope=self.scope)

    def remove(self, key: Any) -> None:
        """
        Remove a button from the side-bar.

        :param key: either a button, index, or name.
        :type key: Any
        :returns None
        """
        self.delete_button(key=key)

    def insert(self, index: int, button: SideBarButton) -> None:
        """
        Place a button at index `index` of the button-list.

        :param index: location index of the new button
        :type index: int
        :param button: a side-bar button object
        :type button: SideBarButton
        """
        if button in self._buttons:
            self._buttons.remove(button)

        self._buttons.insert(check_type(index, int, "index"), button)

    def create_button(self, order: Optional[int] = None, *args, **kwargs) -> SideBarButton:
        """
        Create a side-bar button.

        :param order: Optional input to specify the index of the new button.
        :type order: int
        :return: new side-bar button
        :rtype SideBarButton
        """
        if order is None:
            index = len(self._buttons)
        else:
            index = check_type(order, int, "order")

        button = SideBarButton(side_bar_manager=self, order=index, *args, **kwargs)

        self._buttons.insert(index, button)

        self._update()

        return button

    def add_task_button(
        self,
        activity: "Activity",
        title: Optional[str] = None,
        task_display_mode: Optional[SubprocessDisplayMode] = SubprocessDisplayMode.ACTIVITIES,
        *args,
        **kwargs,
    ) -> SideBarButton:
        """
        Add a side-bar button to a KE-chain activity.

        :param activity: Activity object
        :type activity: Activity
        :param title: Title of the side-bar button, defaults to the activity name
        :type title: str
        :param task_display_mode: for sub-processes, vary the display mode in KE-chain
        :type task_display_mode: SubprocessDisplayMode
        :return: new side-bar button
        :rtype SideBarButton
        """
        from pykechain.models import Activity

        check_type(activity, Activity, "activity")
        check_enum(task_display_mode, SubprocessDisplayMode, "task_display_mode")
        title = check_text(title, "title") or activity.name

        uri = f"{self._scope_uri}/{task_display_mode}/{activity.id}"

        uri_target = (
            URITarget.INTERNAL if activity.scope_id == self.scope.id else URITarget.EXTERNAL
        )

        return self.create_button(uri=uri, uri_target=uri_target, title=title, *args, **kwargs)

    def add_ke_chain_page(
        self, page_name: KEChainPages, title: Optional[str] = None, *args, **kwargs
    ) -> SideBarButton:
        """
        Add a side-bar button to a built-in KE-chain page.

        :param page_name: name of the KE-chain page
        :type page_name: KEChainPages
        :param title: Title of the side-bar button, defaults to the page_name
        :type title: str
        :return: new side-bar button
        :rtype SideBarButton
        """
        page_name = check_enum(page_name, KEChainPages, "page_name")
        title = check_text(title, "title") or KEChainPageLabels[page_name]
        icon = KEChainPageIcons[page_name]
        if "icon" in kwargs:
            icon = kwargs.pop("icon")

        uri = f"{self._scope_uri}/{page_name}"

        return self.create_button(
            uri=uri, uri_target=URITarget.INTERNAL, title=title, icon=icon, *args, **kwargs
        )

    def add_external_button(self, url: str, title: str, *args, **kwargs) -> SideBarButton:
        """
        Add a side-bar button to an external page defined by an URL.

        :param title: title of the button
        :type title: str
        :param url: URL to an external page
        :type url: str
        :return: new side-bar button
        :rtype SideBarButton
        """
        button = self.create_button(
            title=check_text(title, "title"),
            uri=check_url(url),
            uri_target=URITarget.EXTERNAL,
            *args,
            **kwargs,
        )
        return button

    def add_buttons(self, buttons: List[Dict], override_sidebar: bool) -> List[SideBarButton]:
        """
        Create a list of buttons in bulk. Each button is defined by a dict, provided in a sorted list.

        :param buttons: list of dicts
        :type buttons: list
        :param override_sidebar: whether to override the default sidebar menu items.
        :type override_sidebar: bool
        :return: list of SideBarButton objects
        :rtype List[SideBarButton]
        """
        check_list_of_dicts(buttons, "buttons")
        check_type(override_sidebar, bool, "override_sidebar")

        for index, button in enumerate(buttons):
            button = SideBarButton(side_bar_manager=self, order=index, json=button)

            self._buttons.append(button)

        self.override_sidebar = override_sidebar
        self._update()

        return self._buttons

    def delete_button(self, key: Any) -> None:
        """
        Similar to the `remove` method, deletes a button.

        :param key: either a button, index or name
        :return: None
        """
        button = self[key]
        self._buttons.remove(button)
        self._update()

    @property
    def override_sidebar(self) -> bool:
        """
        Flag to indicate whether the original KE-chain side-bar is still shown.

        :return: boolean, True if original side-bar is not visible
        :rtype bool
        """
        return self._override

    @override_sidebar.setter
    def override_sidebar(self, value: bool) -> None:
        """
        Flag to indicate whether the original KE-chain side-bar is still shown.

        :param value: new boolean value
        :type value: bool
        :return: None
        """
        check_type(value, bool, "override_sidebar")
        self._override = value
        self._update()

    def _update(self) -> None:
        """
        Update the side-bar using the scope.options attribute.

        :return: None
        :rtype None
        """
        if self._perform_bulk_creation:
            # Update will proceed during deletion of the manager.
            return

        options = dict(self.scope.options)

        custom_navigation = list()
        for button in self._buttons:
            custom_navigation.append(button.as_dict())

        options.update(
            customNavigation=custom_navigation,
            overrideSideBar=self._override,
        )

        self.scope.options = options



import os
from typing import List
from unittest import TestCase

from pykechain.enums import (
    WidgetTypes,
    ShowColumnTypes,
    FilterType,
    ProgressBarColors,
    Category,
    LinkTargets,
    KEChainPages,
    WidgetTitleValue,
    Alignment,
    ActivityType,
    CardWidgetLinkValue,
    CardWidgetLinkTarget,
    ImageFitValue,
    PropertyType,
    Classification,
    Multiplicity,
    ActivityStatus,
    ActivityClassification,
)
from pykechain.models.widgets.enums import (
    DashboardWidgetShowTasks,
    DashboardWidgetShowScopes,
    TasksAssignmentFilterTypes,
)
from pykechain.exceptions import IllegalArgumentError, NotFoundError
from pykechain.models import Activity, Part
from pykechain.models.widgets import (
    UndefinedWidget,
    HtmlWidget,
    PropertygridWidget,
    AttachmentviewerWidget,
    SupergridWidget,
    FilteredgridWidget,
    TasknavigationbarWidget,
    SignatureWidget,
    ServiceWidget,
    NotebookWidget,
    MulticolumnWidget,
    CardWidget,
    MetapanelWidget,
    ScopeWidget,
    TasksWidget,
)
from pykechain.models.widgets.helpers import _set_title
from pykechain.models.widgets.widget import Widget
from pykechain.models.widgets.widget_models import (
    ServicecardWidget,
    DashboardWidget,
    ScopemembersWidget,
)
from pykechain.models.widgets.widgets_manager import WidgetsManager
from pykechain.utils import slugify_ref, temp_chdir, find
from tests.classes import TestBetamax


class TestSetTitle(TestCase):
    def test_interface(self):
        title_in = "title"
        meta_in = dict()
        output = _set_title(meta_in, title_in)

        self.assertIsInstance(output, tuple)
        self.assertTrue(len(output) == 2)

        meta, title = output

        self.assertIsInstance(meta, dict)
        self.assertIsInstance(title, str)
        self.assertIs(meta_in, meta, msg="Meta must be updated in-place!")
        self.assertIn("showTitleValue", meta)
        self.assertIn("customTitle", meta)

    def test_title(self):
        """Test parser when changing the `title` input."""
        title_in = "title"
        meta, title = _set_title(dict(), title=title_in)
        self.assertEqual(title_in, title)
        self.assertEqual(title_in, meta["customTitle"])
        self.assertEqual(WidgetTitleValue.CUSTOM_TITLE, meta["showTitleValue"])

        title_in = False
        meta, title = _set_title(dict(), title=title_in)
        self.assertIsNone(title)
        self.assertIsNone(meta["customTitle"])
        self.assertEqual(WidgetTitleValue.DEFAULT, meta["showTitleValue"])

        title_in = None
        meta, title = _set_title(dict(), title=title_in)
        self.assertIsNone(title)
        self.assertIsNone(meta["customTitle"])
        self.assertEqual(WidgetTitleValue.NO_TITLE, meta["showTitleValue"])

    def test_show_title_value(self):
        """Test parser when changing the `show_title_value` input."""
        title_in = "title"

        meta, title = _set_title(
            dict(), title=title_in, show_title_value=WidgetTitleValue.NO_TITLE
        )

        self.assertEqual(title_in, title)
        self.assertEqual(title_in, meta["customTitle"])
        self.assertEqual(WidgetTitleValue.NO_TITLE, meta["showTitleValue"])

        meta, title = _set_title(
            dict(), title=title_in, show_title_value=WidgetTitleValue.CUSTOM_TITLE
        )

        self.assertEqual(title_in, title)
        self.assertEqual(title_in, meta["customTitle"])
        self.assertEqual(WidgetTitleValue.CUSTOM_TITLE, meta["showTitleValue"])

        meta, title = _set_title(dict(), title=title_in, show_title_value=WidgetTitleValue.DEFAULT)

        self.assertIsNone(title)
        self.assertIsNone(meta["customTitle"])
        self.assertEqual(WidgetTitleValue.DEFAULT, meta["showTitleValue"])

        with self.assertRaises(
            IllegalArgumentError, msg="Unrecognized show_title_value must be caught!"
        ):
            # noinspection PyTypeChecker
            _set_title(dict(), "title", show_title_value="Maybe")


class TestWidgets(TestBetamax):
    def setUp(self):
        super().setUp()

        self.activity = self.project.activity("Specify wheel diameter")
        self.new_widget = None

    def tearDown(self):
        if self.new_widget:
            self.new_widget.delete()
        super().tearDown()

    def test_retrieve_widgets_in_activity(self):
        widget_set = self.activity.widgets()

        self.assertIsInstance(widget_set, WidgetsManager)

        for w in widget_set:
            with self.subTest(msg=w):
                self.assertIsInstance(w, Widget)
                self.assertEqual(self.activity, w.activity())

    def test_create_widget_in_activity(self):
        self.new_widget = self.client.create_widget(
            title="Test_widget",
            activity=self.activity,
            widget_type=WidgetTypes.UNDEFINED,
            meta={},
            order=100,
        )
        self.assertIsInstance(self.new_widget, UndefinedWidget)

    def test_widget_attributes(self):
        attributes = [
            "_client",
            "_json_data",
            "id",
            "created_at",
            "updated_at",
            "ref",
            "widget_type",
            "title",
            "meta",
            "order",
            "_activity_id",
            "_parent_id",
            "has_subwidgets",
            "_scope_id",
            "progress",
        ]
        first_widget = self.activity.widgets()[0]

        self.assertIsInstance(first_widget, Widget)
        for attribute in attributes:
            with self.subTest(msg=attribute):
                self.assertTrue(hasattr(first_widget, attribute))

    def test_widget_meta_attribute_is_not_None(self):
        first_widget = self.activity.widgets()[0]

        self.assertIsInstance(first_widget, Widget)
        self.assertIsNotNone(first_widget.meta)

    def test_associated_parts(self):
        task = self.project.activity(name="Task - Basic Table")
        grid_widget = list(task.widgets())[0]

        associated_parts = grid_widget.associated_parts()
        self.assertTrue(associated_parts)
        self.assertIsInstance(associated_parts, tuple)
        self.assertTrue(len(associated_parts) == 2)

        models, instances = associated_parts

        self.assertIsInstance(models, list)
        self.assertTrue(all(isinstance(p, Part) for p in models))
        self.assertTrue(all(p.category == Category.MODEL for p in models))
        self.assertTrue(all(p.category == Category.INSTANCE for p in models))

    def test_bulk_update(self):
        parent = self.project.activity(name="Tasks with Widgets")
        task = self.project.activity(name="Task - Form + Tables + Service")
        test_task = task.clone(
            parent=parent,
            name="BULK WIDGET TEST TASK",
        )
        self.new_widget = test_task  # for teardown

        widgets = list(test_task.widgets())[1:]  # remove meta-panel
        self.client.update_widgets(
            widgets=[dict(id=w.id, title=f"widget {i + 1}") for i, w in enumerate(widgets)]
        )

        with self.assertRaises(IllegalArgumentError):
            double_update = [widgets[0], widgets[0], widgets[1]]
            self.client.update_widgets(
                widgets=[
                    dict(id=w.id, title=f"widget {i + 1}") for i, w in enumerate(double_update)
                ]
            )


class TestWidgetManager(TestBetamax):
    def setUp(self):
        super().setUp()
        self.wm = self.project.activity(
            "Task - Form + Tables + Service"
        ).widgets()  # type: WidgetsManager

    def test_widget_manager_creation(self):
        with self.assertRaises(IllegalArgumentError):
            WidgetsManager(widgets=[], activity="Run away!")

    def test_widget_not_found(self):
        with self.assertRaises(NotFoundError):
            widget = self.wm["unladen swallow"]

    def test_activity_has_metapanel_in_widget_manager(self):
        self.assertIsInstance(self.wm, WidgetsManager)
        self.assertTrue(len(self.wm) >= 5)

        self.assertTrue(self.wm[0].widget_type, WidgetTypes.METAPANEL)

    def test_widget_can_be_found_with_uuid_in_widget_manager(self):
        meta_panel = self.wm[0]

        self.assertTrue(meta_panel, WidgetTypes.METAPANEL)
        self.assertEqual(self.wm[meta_panel.id], meta_panel)

    def test_widget_in_widget_manager(self):
        for key in [0, self.wm[0], self.wm[0].id]:
            with self.subTest(msg=f"key: {type(key)} {key}"):
                widget = self.wm[key]
                self.assertIn(widget, self.wm)

    def test_widgetmanager_has_activity_and_client(self):
        self.assertIsNotNone(self.wm._client)
        self.assertIsInstance(self.wm.activity, Activity)
        self.assertIsInstance(self.wm._client, self.client.__class__)


class TestWidgetManagerInActivity(TestBetamax):
    def setUp(self):
        super().setUp()
        self.task = self.project.create_activity(name="widget_test_task")  # type: Activity
        self.wm = self.task.widgets()  # type: WidgetsManager

    def tearDown(self):
        self.task.delete()
        super().tearDown()

    def test_new_widget_using_widget_manager(self):
        self.assertEqual(len(self.wm), 1)
        metapanel = self.wm[0]

        htmlwidget = self.wm.create_widget(
            widget_type=WidgetTypes.HTML, title="Test HTML widget", meta=dict(htmlContent="Hello")
        )

        self.assertIsInstance(metapanel, MetapanelWidget)
        self.assertIsInstance(htmlwidget, HtmlWidget)
        self.assertEqual(len(self.wm), 1 + 1)

    def test_edit_widget_meta(self):
        bike_part = self.project.part("Bike")
        widget = self.wm.add_propertygrid_widget(
            part_instance=bike_part,
            writable_models=bike_part.model().properties,
        )

        widget.meta.update({"showColumns": [ShowColumnTypes.UNIT, ShowColumnTypes.DESCRIPTION]})
        widget.edit(meta=widget.meta)

        live_widget = self.client.widget(pk=widget.id)

        self.assertEqual(widget.meta, live_widget.meta)

    def test_edit_widget_title(self):
        bike_part = self.project.part("Bike")
        widget = self.wm.add_propertygrid_widget(
            part_instance=bike_part,
            writable_models=bike_part.model().properties,
        )

        new_title = "My customly edited title"
        widget.edit(title=new_title)

        live_widget = self.client.widget(pk=widget.id)

        self.assertEqual(new_title, live_widget.meta.get("customTitle"))
        self.assertEqual("Custom title", live_widget.meta.get("showTitleValue"))
        self.assertEqual(new_title, live_widget.title_visible)

    def test_edit_widget_title_is_none(self):
        bike_part = self.project.part("Bike")
        widget = self.wm.add_propertygrid_widget(
            part_instance=bike_part,
            writable_models=bike_part.model().properties,
        )

        widget.edit(title=None)

        live_widget = self.client.widget(pk=widget.id)

        self.assertFalse(live_widget.title)
        self.assertIsNone(live_widget.title_visible)
        self.assertEqual(WidgetTitleValue.NO_TITLE, live_widget.meta.get("showTitleValue"))

        widget.edit(title=False)

        live_widget = self.client.widget(pk=widget.id)

        self.assertFalse(live_widget.title)
        self.assertEqual("Bike", live_widget.title_visible)
        self.assertEqual(WidgetTitleValue.DEFAULT, live_widget.meta.get("showTitleValue"))

    def test_edit_widget_title_and_meta(self):
        bike_part = self.project.part("Bike")
        widget = self.wm.add_propertygrid_widget(
            part_instance=bike_part,
            writable_models=bike_part.model().properties,
        )
        new_title = "New title"
        new_meta = dict(widget.meta)
        new_meta.update({"showColumns": [ShowColumnTypes.UNIT, ShowColumnTypes.DESCRIPTION]})
        widget.edit(
            title=new_title,
            meta=new_meta,
        )

        live_widget = self.client.widget(pk=widget.id)

        self.assertEqual(new_title, live_widget.title_visible)
        self.assertEqual(
            [ShowColumnTypes.UNIT, ShowColumnTypes.DESCRIPTION], live_widget.meta["showColumns"]
        )

    def test_widget_title(self):
        title = "Hidden title"
        bike_part = self.project.part("Bike")
        widget = self.wm.add_propertygrid_widget(
            part_instance=bike_part,
            writable_models=bike_part.model().properties,
            title=title,
            show_title_value=WidgetTitleValue.NO_TITLE,
        )

        self.assertEqual(title, widget.title)
        self.assertNotEqual(widget.title_visible, widget.title)
        self.assertIsNone(widget.title_visible)
        self.assertEqual(slugify_ref(title), widget.ref)
        self.assertEqual(WidgetTitleValue.NO_TITLE, widget.meta.get("showTitleValue"))

    def test_property_grid_with_associations_using_widget_manager(self):

        frame = self.project.part(name="Frame")
        frame_model = self.project.model(name="Frame")

        widget = self.wm.create_widget(
            widget_type=WidgetTypes.PROPERTYGRID,
            title="Frame Property Grid",
            meta=dict(activityId=str(self.task.id), partInstanceId=str(frame.id)),
            writable_models=frame_model.properties,
            readable_models=[],
        )

        self.assertIsInstance(widget, PropertygridWidget)
        self.assertEqual(len(self.wm), 1 + 1)

    def test_add_attachment_widget(self):
        picture_instance = self.project.part("Bike").property("Picture")
        widget = self.wm.add_attachmentviewer_widget(
            attachment_property=picture_instance,
        )

        self.assertIsInstance(widget, AttachmentviewerWidget)
        self.assertEqual(len(self.wm), 1 + 1)
        self.assertEqual("Picture", widget.title_visible)

    def test_add_attachment_widget_with_associations_using_widget_manager(self):
        photo_property = self.project.property("Picture")

        widget = self.wm.add_attachmentviewer_widget(
            title="Attachment Viewer",
            attachment_property=photo_property,
        )

        self.assertEqual("Attachment Viewer", widget.title_visible)
        self.assertIsInstance(widget, AttachmentviewerWidget)
        self.assertEqual(len(self.wm), 1 + 1)

    def test_add_attachment_widget_with_editable_association(self):
        photo_property = self.project.property("Picture")

        widget = self.wm.add_attachmentviewer_widget(
            title="Attachment Viewer",
            attachment_property=photo_property,
            editable=True,
        )

        self.assertIsInstance(widget, AttachmentviewerWidget)
        self.assertEqual(len(self.wm), 1 + 1)

    def test_add_super_grid_widget(self):
        part_model = self.project.model(name="Wheel")
        parent_instance = self.project.part(name="Bike")

        widget = self.wm.add_supergrid_widget(
            part_model=part_model,
            parent_instance=parent_instance,
            edit=False,
            emphasize_edit=True,
            all_readable=True,
            incomplete_rows=True,
        )

        self.assertIsInstance(widget, SupergridWidget)
        self.assertEqual(len(self.wm), 1 + 1)
        self.assertEqual(part_model.name, widget.title_visible)

    def test_add_filtered_grid_widget(self):
        part_model = self.project.model(name="Wheel")
        parent_instance = self.project.part(name="Bike")
        widget = self.wm.add_filteredgrid_widget(
            part_model=part_model,
            parent_instance=parent_instance,
            edit=True,
            # sort_property=part_model.property(name='Diameter'),
            emphasize_edit=True,
            all_writable=True,
            collapse_filters=False,
        )

        self.assertIsInstance(widget, FilteredgridWidget)
        self.assertEqual(len(self.wm), 1 + 1)
        self.assertEqual(part_model.name, widget.title_visible)

    def test_add_filtered_grid_widget_with_prefilters_and_excluded_propmodels(self):
        part_model = self.project.model(name="Wheel")
        parent_instance = self.project.part(name="Bike")
        excluded_propmodels = [part_model.property(name="Spokes")]
        prefilters = dict(
            property_models=[part_model.property(name="Diameter")],
            values=[66],
            filters_type=[FilterType.LOWER_THAN_EQUAL],
        )
        widget = self.wm.add_filteredgrid_widget(
            part_model=part_model,
            parent_instance=parent_instance,
            custom_height=500,
            edit=True,
            sort_name=True,
            emphasize_edit=True,
            all_writable=True,
            collapse_filters=False,
            prefilters=prefilters,
            excluded_propmodels=excluded_propmodels,
        )

        self.assertIsInstance(widget, FilteredgridWidget)
        self.assertEqual(len(self.wm), 1 + 1)

    def test_add_propertygrid_widget(self):
        bike_part = self.project.part(name="Bike")
        widget = self.wm.add_propertygrid_widget(
            part_instance=bike_part,
            show_headers=False,
            show_columns=[ShowColumnTypes.UNIT],
            readable_models=bike_part.model().properties[:2],
            writable_models=bike_part.model().properties[3:],
        )

        self.assertIsInstance(widget, PropertygridWidget)
        self.assertEqual(len(self.wm), 1 + 1)
        self.assertEqual(bike_part.name, widget.title_visible)

    def test_add_signature_widget(self):
        bike_part = self.project.part(name="Bike")
        picture = bike_part.property(name="Picture")

        widget1 = self.wm.add_signature_widget(
            attachment_property=picture,
            title="Yes, my precious",
            custom_undo_button_text="Remove za widget",
            custom_button_text="Sign za widget, plz",
            show_name_and_date=False,
        )
        widget2 = self.wm.add_signature_widget(
            attachment_property=picture,
            title=False,
            custom_undo_button_text=False,
            custom_button_text=False,
        )
        widget3 = self.wm.add_signature_widget(attachment_property=picture)
        self.assertIsInstance(widget1, SignatureWidget)
        self.assertIsInstance(widget2, SignatureWidget)
        self.assertIsInstance(widget3, SignatureWidget)
        self.assertEqual(len(self.wm), 1 + 3)

        with self.assertRaises(IllegalArgumentError):
            self.wm.add_signature_widget(attachment_property="Failed script")

    def test_add_card_widget(self):
        # setUp
        bike_part = self.project.part(name="Bike")
        picture = bike_part.property(name="Picture")
        process = self.project.activities(activity_type=ActivityType.PROCESS)[0]

        widget1 = self.wm.add_card_widget(description="Some description")
        widget2 = self.wm.add_card_widget(
            image=picture,
            title=False,
            image_fit=ImageFitValue.COVER,
            link=self.task.id,
            link_target=LinkTargets.SAME_TAB,
        )
        widget3 = self.wm.add_card_widget(
            title="Tree view", description="Process opens in tree view", link=process
        )

        # testing
        self.assertEqual(len(self.wm), 1 + 3)

        self.assertEqual(self.project.name, widget1.title_visible)
        self.assertIsInstance(widget1, CardWidget)
        self.assertIsNone(widget1.meta.get("customLink"))

        self.assertIsInstance(widget2, CardWidget)
        self.assertEqual(ImageFitValue.COVER, widget2.meta.get("imageFit"))
        self.assertEqual(CardWidgetLinkTarget.SAME_TAB, widget2.meta.get("linkTarget"))

        self.assertIsInstance(widget3, CardWidget)
        self.assertEqual(CardWidgetLinkValue.TREE_VIEW, widget3.meta.get("showLinkValue"))

        with self.assertRaises(IllegalArgumentError):
            self.wm.add_card_widget(title=12)

        with self.assertRaises(IllegalArgumentError):
            self.wm.add_card_widget(description=bike_part)

        with self.assertRaises(IllegalArgumentError):
            self.wm.add_card_widget(image="this should not work")

        with self.assertRaises(IllegalArgumentError):
            self.wm.add_card_widget(link=self.task, link_target="_somewhere")

        with self.assertRaises(IllegalArgumentError):
            # noinspection PyTypeChecker
            self.wm.add_card_widget(image_fit=3)

    def test_add_card_widget_ke_chain_pages(self):
        for native_page_name in KEChainPages.values():
            with self.subTest(msg=f"Page {native_page_name}"):
                card_widget = self.wm.add_card_widget(
                    title=native_page_name, link=native_page_name
                )
                self.assertIsInstance(card_widget, CardWidget)

        self.assertEqual(
            len(self.wm),
            9,
            msg=(
                "New KE-chain page has been added to the Enum, "
                "check if the mapping dicts in enums.py need updating too!"
            ),
        )

    def test_add_service_widget(self):
        service_gears_successful = self.project.service("Service Gears - Successful")

        widget1 = self.wm.add_service_widget(service=service_gears_successful)
        widget2 = self.wm.add_service_widget(
            service=service_gears_successful,
            title=None,
            custom_button_text="Run this script (Custom!) and no title",
            emphasize_run=True,
            download_log=True,
        )
        widget3 = self.wm.add_service_widget(
            service=service_gears_successful,
            title="Also a custom title, but no log",
            custom_button_text="Run this script (Custom!)",
            emphasize_run=False,
            show_log=False,
        )

        self.assertIsInstance(widget1, ServiceWidget)
        self.assertIsInstance(widget2, ServiceWidget)
        self.assertIsInstance(widget3, ServiceWidget)
        self.assertEqual(len(self.wm), 1 + 3)
        self.assertEqual(service_gears_successful.name, widget1.title_visible)

    def test_add_html_widget(self):
        widget_1 = self.wm.add_html_widget(
            html="Or is this just fantasy?", title="Is this the real life?"
        )

        self.assertIsInstance(widget_1, HtmlWidget)
        self.assertEqual(len(self.wm), 1 + 1)

        widget_2 = self.wm.add_html_widget(
            title="Caught in a landslide",
            show_title_value=WidgetTitleValue.NO_TITLE,
            html="No escape from reality",
        )

        self.assertEqual(widget_2.ref, slugify_ref("Caught in a landslide"))

    def test_add_metapanel_widget(self):
        widget = self.wm.add_metapanel_widget(
            show_all=True, show_progress=True, show_progressbar=True
        )

        self.assertEqual(len(self.wm), 1 + 1)
        self.assertIsNone(widget.title_visible)

    def test_add_metapanel_with_progress_settings(self):
        progress_bar = dict(
            height=15, showProgressText=False, colorCompleted=ProgressBarColors.RED
        )
        self.wm.add_metapanel_widget(
            show_all=False, show_progress=False, show_progressbar=True, progress_bar=progress_bar
        )
        self.assertEqual(len(self.wm), 1 + 1)

    def test_add_progress_widget(self):
        self.wm.add_progress_widget(custom_height=35, show_progress_text=False)
        self.assertEqual(len(self.wm), 1 + 1)

    def test_delete_all_widgets(self):
        """Delete all widgets from an activity"""
        self.assertEqual(len(self.wm), 1)

        self.wm.delete_all_widgets()

        self.assertEqual(len(self.wm), 0)
        self.assertEqual(len(self.task.widgets()), 0)

    def test_delete_widget(self):
        """Delete single widget from an activity"""
        self.assertEqual(len(self.wm), 1)

        self.wm.delete_widget(0)

        self.assertEqual(len(self.wm), 0)
        self.assertEqual(len(self.task.widgets()), 0)

    def test_delete_widget_stand_alone(self):
        """Delete a widget by itself and check the manager"""
        widget_one = self.wm[0]
        widget_one.delete()

        self.assertNotIn(widget_one, self.wm)

    def test_add_notebook_widget(self):
        notebook = self.project.service(name="Service Gears - Successful")
        widget1 = self.wm.add_notebook_widget(notebook=notebook, title=False)

        widget2 = self.wm.add_notebook_widget(notebook=notebook, title="With custom title")

        widget3 = self.wm.add_notebook_widget(
            notebook=notebook.id,
            title="With no padding and custom height",
            customHeight=400,
            noPadding=True,
        )

        self.assertIsInstance(widget1, NotebookWidget)
        self.assertIsInstance(widget2, NotebookWidget)
        self.assertIsInstance(widget3, NotebookWidget)
        self.assertEqual(len(self.wm), 1 + 3)

        with self.assertRaises(IllegalArgumentError):
            self.wm.add_notebook_widget(notebook="This will raise an error")

    def test_add_multicolumn_widget(self):
        bike_part = self.project.part("Bike")
        picture_instance = bike_part.property("Picture")
        picture_model = picture_instance.model()
        multi_column_widget = self.wm.add_multicolumn_widget(
            title="Multi column Grid + Attachment"
        )

        widget1 = self.wm.add_propertygrid_widget(
            part_instance=bike_part,
            writable_models=[picture_model],
            parent_widget=multi_column_widget,
        )

        widget2 = self.wm.add_attachmentviewer_widget(
            attachment_property=picture_instance, parent_widget=multi_column_widget.id
        )

        self.assertIsInstance(multi_column_widget, MulticolumnWidget)
        self.assertIsInstance(widget1, PropertygridWidget)
        self.assertIsInstance(widget2, AttachmentviewerWidget)
        self.assertEqual(len(self.wm), 1 + 3)

    def test_parent(self):
        multi_column_widget = self.wm.add_multicolumn_widget(title="Multi column widget")

        child_widget = self.wm.add_html_widget(
            title="Text widget", html="Text", parent_widget=multi_column_widget
        )

        parent_widget = child_widget.parent()

        self.assertEqual(multi_column_widget, parent_widget)

        with self.assertRaises(NotFoundError):
            parent_widget.parent()

    def test_add_scope_widget(self):
        scope_widget = self.wm.add_scope_widget()

        self.assertIsInstance(scope_widget, ScopeWidget)

    def test_scope_widget_invalid_inputs(self):
        for inputs in [
            dict(team="5"),
            dict(add=1),
            dict(edit=0),
            dict(emphasize_add="True"),
            dict(emphasize_edit=1.0),
            dict(show_columns="All", show_all_columns=False),
            dict(show_columns=["Any"], show_all_columns=False),
            dict(show_all_columns=0),
            dict(page_size=0),
            dict(tags="one"),
            dict(tags=["one", 2, "three"]),
            dict(sorted_column="Project name"),
            dict(sorted_direction="Alphabetical"),
            dict(active_filter="Active"),
            dict(search_filter="Project"),
        ]:
            with self.subTest(msg=inputs):
                with self.assertRaises(IllegalArgumentError):
                    self.wm.add_scope_widget(**inputs)

    def test_insert_widget(self):
        bike_part = self.project.part("Bike")
        w0 = self.wm[0]  # meta panel

        w1, w2, w3 = (
            self.wm.add_propertygrid_widget(
                part_instance=bike_part,
                writable_models=bike_part.model().properties,
                title="Original widget {i} (w{i})".format(i=i + 1),
            )
            for i in range(3)
        )

        # if widget order is `[w0,w1,w2]` and inserting `w3` at index 1 (before Widget1);
        #           index:     0 ^1  2
        # the list will be `[w0,w3,w1,w2]`
        self.wm.insert(1, w3)

        self.assertTrue(w0.widget_type, WidgetTypes.METAPANEL)
        self.assertEqual(self.wm[w0.id].order, 0)
        self.assertEqual(self.wm[w3.id].order, 1)
        self.assertEqual(self.wm[w1.id].order, 2)
        self.assertEqual(self.wm[w2.id].order, 3)

        added_widget = self.wm.add_propertygrid_widget(
            part_instance=bike_part,
            writable_models=bike_part.model().properties,
            title="Widget Finally Positioned Under Metapanel",
            order=1,
        )

        self.assertEqual(added_widget.order, 1)
        self.assertEqual(self.wm[w3.id].order, 2)
        self.assertEqual(self.wm[w1.id].order, 3)
        self.assertEqual(self.wm[w2.id].order, 4)

    def test_create_widgets(self):
        # setUp
        new_widgets = self.wm.create_widgets(
            widgets=[
                dict(
                    widget_type=WidgetTypes.HTML,
                    title="A new text widget",
                    meta=dict(htmlContent="This is HTML text."),
                ),
                dict(
                    widget_type=WidgetTypes.HTML,
                    title="Another HTML widget",
                    meta=dict(htmlContent="They keep on multiplying."),
                ),
            ]
        )

        # testing
        self.assertIsInstance(new_widgets, list)
        self.assertTrue(all(isinstance(w, Widget) for w in new_widgets))
        self.assertEqual(len(new_widgets), 2)
        self.assertIn(new_widgets[0], self.wm._widgets)

        # tearDown
        [w.delete() for w in new_widgets]

    def test_add_service_card_widget(self):
        # setUp
        service_gears_successful = self.project.service("Service Gears - Successful")
        bike_part = self.project.part(name="Bike")
        picture = bike_part.property(name="Picture")
        title = "Service card title"
        description = "The description"

        widget = self.wm.add_service_card_widget(
            service=service_gears_successful,
            image=picture,
            title=title,
            description=description,
            custom_button_text=False,
            link=KEChainPages.EXPLORER,
            link_target=LinkTargets.SAME_TAB,
            image_fit=ImageFitValue.COVER,
        )

        self.assertEqual(len(self.wm), 1 + 1)

        self.assertIsInstance(widget, ServicecardWidget)
        self.assertEqual(widget.title, title)
        self.assertEqual(widget.meta.get("customText"), service_gears_successful.name)
        self.assertEqual(widget.meta.get("customDescription"), description)
        self.assertIn(picture.id, widget.meta.get("customImage"))
        self.assertEqual(ImageFitValue.COVER, widget.meta.get("imageFit"))
        self.assertEqual(CardWidgetLinkTarget.SAME_TAB, widget.meta.get("linkTarget"))

    def test_add_dashboard_widget(self):
        # setUp
        title_widget_1 = "Dashboard widget"
        title_widget_2 = "Tagged projects widget"

        widget_current_project = self.wm.add_dashboard_widget(
            title=title_widget_1, show_assignees=True
        )
        widget_tagged_projects = self.wm.add_dashboard_widget(
            title=title_widget_2,
            source_scopes_tags=["catalog-scope"],
            show_tasks=[
                DashboardWidgetShowTasks.UNASSIGNED_TASKS,
                DashboardWidgetShowTasks.CLOSED_TASKS,
            ],
            show_scopes=[DashboardWidgetShowScopes.CLOSED_SCOPES],
            show_assignees=False,
        )

        # testing
        self.assertIsInstance(widget_current_project, DashboardWidget)
        self.assertIsInstance(widget_tagged_projects, DashboardWidget)

        self.assertEqual(widget_current_project.title, title_widget_1)
        self.assertEqual(widget_tagged_projects.title, title_widget_2)

        self.assertTrue(
            all(elem["selected"] for elem in widget_current_project.meta["showNumbers"])
        )
        self.assertTrue(
            all(elem["selected"] for elem in widget_current_project.meta["showNumbersProjects"])
        )

        self.assertTrue(
            any(elem["selected"] is False for elem in widget_tagged_projects.meta["showNumbers"])
        )
        self.assertTrue(
            any(
                elem["selected"] is False
                for elem in widget_tagged_projects.meta["showNumbersProjects"]
            )
        )

        self.assertTrue(widget_current_project.meta["showAssignees"])
        self.assertFalse(widget_tagged_projects.meta["showAssignees"])

    def test_add_tasks_widget(self):
        tasks_widget = self.wm.add_tasks_widget()

        self.assertIsInstance(tasks_widget, TasksWidget)

    def test_add_tasks_widget_with_filters(self):
        tasks_widget = self.wm.add_tasks_widget(
            parent_activity=self.wm.activity.parent(),
            status_filter=ActivityStatus.OPEN,
            assigned_filter=TasksAssignmentFilterTypes.FILTER_ASSIGNED_TO_USER,
            activity_type_filter=ActivityType.TASK,
            classification_filter=ActivityClassification.CATALOG,
            tags_filter=["One", "Two", "Five"],
        )

        self.assertIsInstance(tasks_widget, TasksWidget)

    def test_add_scopemembers_widget(self):
        scope_members_widget = self.wm.add_scopemembers_widget()

        self.assertIsInstance(scope_members_widget, ScopemembersWidget)


class TestWidgetManagerWeatherWidget(TestBetamax):
    def setUp(self):
        super().setUp()
        self.task = self.project.create_activity(name="widget_test_task")  # type: Activity
        self.wm = self.task.widgets()  # type: WidgetsManager

        catalog_root_model = self.project.part(
            name="Catalog", classification=Classification.CATALOG, category=Category.MODEL
        )
        self.part_model_with_weather_prop = self.project.create_model_with_properties(
            parent=catalog_root_model,
            name="___TEST PART",
            multiplicity=Multiplicity.ONE,
            properties_fvalues=[dict(name="weather", property_type=PropertyType.WEATHER_VALUE)],
        )
        self.weather_prop_instance = self.part_model_with_weather_prop.instances()[0].property(
            "weather"
        )

    def tearDown(self):
        self.weather_widget.delete()
        self.part_model_with_weather_prop.delete()
        self.task.delete()
        super().tearDown()

    def test_weather_widget(self):
        """Testing the weather widget."""
        self.weather_widget = self.wm.add_weather_widget(
            weather_property=self.weather_prop_instance,
        )


class TestWidgetNavigationBarWidget(TestBetamax):
    def setUp(self):
        super().setUp()
        self.task = self.project.create_activity(name="widget_test_task")  # type: Activity
        self.wm = self.task.widgets()  # type: WidgetsManager

        self.frame = self.project.part(name="Frame")
        self.frame_model = self.project.model(name="Frame")

        activity_1 = self.project.activity("Task - Basic Table")
        activity_2 = self.project.activity("Task - Form")
        activity_3 = self.project.activity("Task - Service Increase Gears")

        self.nav_bar_config = [
            {"activityId": activity_1, "emphasize": True},
            {"activityId": activity_2, "customText": "Hit me baby!"},
            {"activityId": activity_3},
        ]

    def tearDown(self):
        self.task.delete()
        super().tearDown()

    def test_add_navbar_widget(self):
        widget = self.wm.add_tasknavigationbar_widget(
            activities=self.nav_bar_config, title="Navbar", alignment=Alignment.LEFT
        )

        self.assertIsInstance(widget, TasknavigationbarWidget)
        self.assertEqual(len(self.wm), 1 + 1)

    def test_add_navbar_widget_in_place(self):
        """
        Test whether the inputs to the function are updated in-place (incorrect) or not (correct).
        For example, activity objects are replaced by their IDs and customText is converted to a string.
        """
        self.nav_bar_config[0]["customText"] = False

        import copy

        original_bar = copy.deepcopy(self.nav_bar_config)

        self.wm.add_tasknavigationbar_widget(
            activities=self.nav_bar_config,
            alignment=Alignment.LEFT,
        )

        self.assertEqual(original_bar, self.nav_bar_config)

    def test_add_navbar_widget_incorrect_keys(self):
        self.nav_bar_config[0]["emphasizeButton"] = True

        with self.assertRaises(IllegalArgumentError):
            self.wm.add_tasknavigationbar_widget(
                activities=self.nav_bar_config,
                alignment=Alignment.LEFT,
            )

    def test_add_navbar_external_link(self):
        link = "https://www.google.com"

        button_1 = self.nav_bar_config[0]
        button_1.pop("activityId")
        button_1["link"] = link

        widget = self.wm.add_tasknavigationbar_widget(
            activities=self.nav_bar_config,
            alignment=Alignment.LEFT,
        )

        self.assertEqual(widget.meta["taskButtons"][0]["link"], link)

    def test_add_navbar_disabled_button(self):
        button_1 = self.nav_bar_config[0]
        button_1["isDisabled"] = True

        widget = self.wm.add_tasknavigationbar_widget(
            activities=self.nav_bar_config,
            alignment=Alignment.LEFT,
        )

        self.assertTrue(widget.meta["taskButtons"][0]["isDisabled"])


class TestWidgetsCopyMove(TestBetamax):
    def setUp(self):
        super().setUp()
        self.task = self.project.create_activity(name="widget_test_task")  # type: Activity
        self.task_2 = self.project.create_activity(name="test_copy_widget")  # type: Activity

    def tearDown(self):
        self.task.delete()
        self.task_2.delete()
        super().tearDown()

    def test_copy_widget(self):
        # setUp
        title = "Widget to copy"
        bike_part = self.project.part("Bike")
        widget_manager = self.task.widgets()  # type: WidgetsManager
        widget_1 = widget_manager.add_propertygrid_widget(
            part_instance=bike_part, all_writable=True, title=title
        )
        widget_1.copy(target_activity=self.task_2, order=0)
        widget_manager_2 = self.task_2.widgets()  # type: WidgetsManager
        associated_model = widget_manager_2[0].parts(category=Category.MODEL)[0]

        # testing
        self.assertEqual(widget_manager_2[0].widget_type, WidgetTypes.PROPERTYGRID)
        self.assertEqual(widget_manager_2[0].title, title)
        self.assertTrue(all(prop.output for prop in associated_model.properties))
        self.assertEqual(len(self.task.widgets()), 2)

    def test_copy_widget_with_wrong_target(self):
        # setUp
        title = "Widget to copy"
        bike_part = self.project.part("Bike")
        self.wm = self.task.widgets()  # type: WidgetsManager
        widget_1 = self.wm.add_propertygrid_widget(
            part_instance=bike_part, all_writable=True, title=title
        )

        # testing
        with self.assertRaises(IllegalArgumentError):
            widget_1.copy(target_activity=self.project, order=0)

    def test_move_widget(self):
        # setUp
        title = "Widget to move"
        bike_part = self.project.part("Bike")
        widget_manager = self.task.widgets()  # type: WidgetsManager
        widget_1 = widget_manager.add_propertygrid_widget(
            part_instance=bike_part, all_writable=True, title=title
        )
        widget_1.move(target_activity=self.task_2, order=0)
        widget_manager_2 = self.task_2.widgets()  # type: WidgetsManager
        associated_model = widget_manager_2[0].parts(category=Category.MODEL)[0]

        # testing
        self.assertEqual(widget_manager_2[0].widget_type, WidgetTypes.PROPERTYGRID)
        self.assertEqual(widget_manager_2[0].title, title)
        self.assertTrue(all(prop.output for prop in associated_model.properties))
        self.assertEqual(1, len(self.task.widgets()))


class TestWidgetDownloadAsExcel(TestBetamax):
    def setUp(self):
        super().setUp()
        task = self.project.activity(name="Task - Form + Tables + Service")
        all_widgets = task.widgets()  # type: List[Widget]

        self.grid_widget = find(all_widgets, lambda w: w.widget_type == WidgetTypes.FILTEREDGRID)
        self.form_widget = find(all_widgets, lambda w: w.widget_type == WidgetTypes.PROPERTYGRID)

    def tearDown(self):
        super().tearDown()

    def test_download_as_excel(self):
        file_name = "My excel file"

        with temp_chdir() as target_dir:

            path = self.grid_widget.download_as_excel(
                target_dir=target_dir,
                file_name=file_name,
            )

            self.assertTrue(os.path.exists(path))
            self.assertIn(slugify_ref(file_name), path)

    def test_timezone_aware(self):
        file_name = "My excel file.xls"
        user = self.client.user("testuser")
        timezone_definition = user._json_data["timezone"]

        self.assertEqual("Europe/Amsterdam", timezone_definition)

        with temp_chdir() as target_dir:
            path = self.grid_widget.download_as_excel(
                target_dir=target_dir,
                file_name=file_name,
                user=user,
            )

            self.assertTrue(os.path.exists(path))

    # noinspection PyTypeChecker
    def test_invalid_inputs(self):
        with self.assertRaises(IllegalArgumentError):
            self.form_widget.download_as_excel()

        with self.assertRaises(IllegalArgumentError):
            self.grid_widget.download_as_excel(target_dir=3)

        with self.assertRaises(IllegalArgumentError):
            self.grid_widget.download_as_excel(file_name=15.3)

        with self.assertRaises(IllegalArgumentError):
            self.grid_widget.download_as_excel(user="Testuser")



from copy import deepcopy

widget_meta_schema = {
    "$schema": "http://json-schema.org/draft-06/schema#",
    "title": "Toplevel Widget Meta schema",
    "type": "object",
    "additionalProperties": False,
    "properties": {},
    "definitions": {
        "uuidString": {
            "type": "string",
            "pattern": (
                "^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$"
            ),
        },
        "nullUuidString": {
            "type": ["string", "null"],
            "pattern": (
                "^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$"
            ),
        },
        "nullString": {"type": ["string", "null"]},
        "positiveInteger": {"type": ["integer", "null", "string"], "minimum": 0},
        "booleanNull": {"type": ["boolean", "null"]},
    },
}

undefined_meta_schema = deepcopy(widget_meta_schema)
undefined_meta_schema.update({"additionalProperties": True})



"""All pykechain configuration constants will be listed here."""
#
# Configuration of async download of activity pdf exports
#
ASYNC_REFRESH_INTERVAL = 2  # seconds
ASYNC_TIMEOUT_LIMIT = 100  # seconds

#
# Configuration of the retry options for the client requests based on `urlib3.utils.Retry`.
#

# How many connection-related errors to retry on.
# These are errors raised before the request is sent to the remote server,
# which we assume has not triggered the server to process the request.
RETRY_ON_CONNECTION_ERRORS = 3  # times

# How many times to retry on read errors.
# These errors are raised after the request was sent to the server, so the request may have side-effects.
RETRY_ON_READ_ERRORS = 2

# How many redirects to perform. Limit this to avoid infinite redirect loops.
# A redirect is a HTTP response with a status code 301, 302, 303, 307 or 30
RETRY_ON_REDIRECT_ERRORS = 1

# Total number of retries to allow. Takes precedence over other counts
RETRY_TOTAL = 10

# A backoff factor to apply between attempts after the second try (most errors are resolved
# immediately by a second try without a delay). urllib3 will sleep for:
# {backoff factor} * (2 ** ({number of total retries} - 1)) seconds.
# If the backoff_factor is 0.1, then sleep() will sleep for [0.0s, 0.2s, 0.4s, …] between retries.
RETRY_BACKOFF_FACTOR = 0.8

# Batching of parts when a large number of parts are requested at once
PARTS_BATCH_LIMIT = 100  # number of parts

#
# API Paths and API Extra Parameters
#
API_PATH = {
    "activities": "api/activities.json",
    "activity": "api/activities/{activity_id}.json",
    "activity_export": "api/activities/{activity_id}/export",
    "activity_move": "api/activities/{activity_id}/move_activity",
    "activities_bulk_clone": "api/activities/bulk_clone",
    "activities_bulk_update": "api/activities/bulk_update",
    "widgets_config": "api/widget_config.json",
    "widget_config": "api/widget_config/{widget_config_id}.json",
    "services": "api/services.json",
    "service": "api/services/{service_id}.json",
    "service_execute": "api/services/{service_id}/execute",
    "service_upload": "api/services/{service_id}/upload",
    "service_download": "api/services/{service_id}/download",
    "service_executions": "api/service_executions.json",
    "service_execution": "api/service_executions/{service_execution_id}.json",
    "service_execution_terminate": "api/service_executions/{service_execution_id}/terminate",
    "service_execution_notebook_url": "api/service_executions/{service_execution_id}/notebook_url",
    "service_execution_log": "api/service_executions/{service_execution_id}/log",
    "users": "api/users.json",
    "user_current": "api/users/get_current_user",
    "teams": "api/teams.json",
    "team": "api/teams/{team_id}.json",
    "team_add_members": "api/teams/{team_id}/add_members",
    "team_remove_members": "api/teams/{team_id}/remove_members",
    "versions": "api/versions.json",
    "associations": "api/associations.json",
    # PIM3
    "scope": "api/v3/scopes/{scope_id}.json",
    "scope_add_member": "api/v3/scopes/{scope_id}/add_member",
    "scope_remove_member": "api/v3/scopes/{scope_id}/remove_member",
    "scope_add_manager": "api/v3/scopes/{scope_id}/add_manager",
    "scope_remove_manager": "api/v3/scopes/{scope_id}/remove_manager",
    "scope_add_leadmember": "api/v3/scopes/{scope_id}/add_leadmember",
    "scope_remove_leadmember": "api/v3/scopes/{scope_id}/remove_leadmember",
    "scope_add_supervisor": "api/v3/scopes/{scope_id}/add_supervisor",
    "scope_remove_supervisor": "api/v3/scopes/{scope_id}/remove_supervisor",
    "scopes": "api/v3/scopes.json",
    "scopes_clone": "api/v3/scopes/clone",
    "parts": "api/v3/parts.json",
    "parts_new_instance": "api/v3/parts/new_instance",
    "parts_create_child_model": "api/v3/parts/create_child_model",
    "parts_create_proxy_model": "api/v3/parts/create_proxy_model",
    "parts_clone_model": "api/v3/parts/clone_model",
    "parts_clone_instance": "api/v3/parts/clone_instance",
    "parts_bulk_create": "/api/v3/parts/bulk_create_part_instances",
    "parts_bulk_delete": "/api/v3/parts/bulk_delete_part_instances",
    "parts_export": "api/v3/parts/export",
    "part": "api/v3/parts/{part_id}.json",
    "properties": "api/v3/properties.json",
    "properties_bulk_update": "api/v3/properties/bulk_update",
    "properties_create_model": "api/v3/properties/create_model",
    "property": "api/v3/properties/{property_id}.json",
    "property_upload": "api/v3/properties/{property_id}/upload",
    "property_download": "api/v3/properties/{property_id}/download",
    "widgets": "api/widgets.json",
    "widget": "api/widgets/{widget_id}.json",
    "widget_clear_associations": "api/widgets/{widget_id}/clear_associations.json",
    "widget_remove_associations": "api/widgets/{widget_id}/remove_associations.json",
    "widgets_update_associations": "api/widgets/bulk_update_associations.json",
    "widget_update_associations": "api/widgets/{widget_id}/update_associations.json",
    "widgets_set_associations": "api/widgets/bulk_set_associations.json",
    "widget_set_associations": "api/widgets/{widget_id}/set_associations.json",
    "widgets_bulk_create": "api/widgets/bulk_create",
    "widgets_bulk_delete": "api/widgets/bulk_delete",
    "widgets_bulk_update": "api/widgets/bulk_update",
    "widgets_schemas": "api/widgets/schemas",
    "notifications": "api/v3/notifications.json",
    "notification": "api/v3/notifications/{notification_id}.json",
    "notification_share_activity_link": "api/v3/notifications/share_activity_link",
    "notification_share_activity_pdf": "api/v3/notifications/share_activity_pdf",
    "banners": "api/v3/banners.json",
    "banner": "api/v3/banners/{banner_id}.json",
    "banner_active": "api/v3/banners/active.json",
    "expiring_downloads": "api/downloads.json",
    "expiring_download": "api/downloads/{download_id}.json",
    "expiring_download_download": "api/downloads/{download_id}/download",
    "expiring_download_upload": "api/downloads/{download_id}/upload",
    "contexts": "api/v3/contexts/contexts.json",
    "context": "api/v3/contexts/contexts/{context_id}.json",
    "context_link_activities": "api/v3/contexts/contexts/{context_id}/link_activities",
    "context_unlink_activities": "api/v3/contexts/contexts/{context_id}/unlink_activities",
    # 'feature_collections': 'api/v3/contexts/feature_collections.json',
    # 'feature_collection': 'api/v3/contexts/feature_collections/{context_id}.json',
    # 'time_periods': 'api/v3/contexts/time_periods.json',
    # 'time_period': 'api/v3/contexts/time_periods/{context_id}.json'
}

API_QUERY_PARAM_ALL_FIELDS = {"fields": "__all__"}
API_EXTRA_PARAMS = {
    "activity": {
        "fields": ",".join(
            [
                "id",
                "name",
                "ref",
                "description",
                "created_at",
                "updated_at",
                "activity_type",
                "classification",
                "tags",
                "progress",
                "assignees_ids",
                "start_date",
                "due_date",
                "status",
                "parent_id",
                "scope_id",
                "customization",
                "activity_options",
            ]
        )
    },
    "activities": {
        "fields": ",".join(
            [
                "id",
                "name",
                "ref",
                "description",
                "created_at",
                "updated_at",
                "activity_type",
                "classification",
                "tags",
                "progress",
                "assignees_ids",
                "start_date",
                "due_date",
                "status",
                "parent_id",
                "scope_id",
                "customization",
                "activity_options",
            ]
        )
    },
    "banner": {
        "fields": ",".join(
            [
                "id",
                "text",
                "icon",
                "is_active",
                "active_from",
                "active_until",
                "url",
                "created_at",
                "updated_at",
            ]
        )
    },
    "banners": {
        "fields": ",".join(
            [
                "id",
                "text",
                "icon",
                "is_active",
                "active_from",
                "active_until",
                "url",
                "created_at",
                "updated_at",
            ]
        )
    },
    "scope": {
        "fields": ",".join(
            [
                "id",
                "name",
                "ref",
                "text",
                "created_at",
                "updated_at",
                "start_date",
                "due_date",
                "status",
                "category",
                "progress",
                "members",
                "team",
                "tags",
                "scope_options",
                "team_id_name",
                "workflow_root_id",
                "catalog_root_id",
                "app_root_id",
                "product_model_id",
                "product_instance_id",
                "catalog_model_id",
                "catalog_instance_id",
            ]
        )
    },
    "scopes": {
        "fields": ",".join(
            [
                "id",
                "name",
                "ref",
                "text",
                "created_at",
                "updated_at",
                "start_date",
                "due_date",
                "status",
                "category",
                "progress",
                "members",
                "team",
                "tags",
                "scope_options",
                "team_id_name",
                "workflow_root_id",
                "catalog_root_id",
                "app_root_id",
                "product_model_id",
                "product_instance_id",
                "catalog_model_id",
                "catalog_instance_id",
            ]
        )
    },
    "part": {
        "fields": ",".join(
            [
                "id",
                "name",
                "ref",
                "description",
                "created_at",
                "updated_at",
                "properties",
                "category",
                "classification",
                "parent_id",
                "multiplicity",
                "value_options",
                "property_type",
                "value",
                "output",
                "order",
                "part_id",
                "scope_id",
                "model_id",
                "proxy_source_id_name",
                "unit",
            ]
        )
    },
    "parts": {
        "fields": ",".join(
            [
                "id",
                "name",
                "ref",
                "description",
                "created_at",
                "updated_at",
                "properties",
                "category",
                "classification",
                "parent_id",
                "multiplicity",
                "value_options",
                "property_type",
                "value",
                "output",
                "order",
                "part_id",
                "scope_id",
                "model_id",
                "proxy_source_id_name",
                "unit",
            ]
        )
    },
    "properties": {
        "fields": ",".join(
            [
                "id",
                "name",
                "ref",
                "created_at",
                "updated_at",
                "model_id",
                "part_id",
                "order",
                "scope_id",
                "category",
                "property_type",
                "value",
                "value_options",
                "output",
                "description",
                "unit",
            ]
        )
    },
    "property": {
        "fields": ",".join(
            [
                "id",
                "name",
                "ref",
                "created_at",
                "updated_at",
                "model_id",
                "part_id",
                "order",
                "scope_id",
                "category",
                "property_type",
                "value",
                "value_options",
                "output",
                "description",
                "unit",
            ]
        )
    },
    "service": {
        "fields": ",".join(
            [
                "id",
                "name",
                "ref",
                "created_at",
                "updated_at",
                "script_version",
                "script_type",
                "script_file_name",
                "description",
                "env_version",
                "scope",
                "run_as",
                "trusted",
                "verified_on",
                "verification_results",
            ]
        )
    },
    "services": {
        "fields": ",".join(
            [
                "id",
                "name",
                "ref",
                "created_at",
                "updated_at",
                "script_version",
                "script_type",
                "script_file_name",
                "description",
                "env_version",
                "scope",
                "run_as",
                "trusted",
                "verified_on",
                "verification_results",
            ]
        )
    },
    "widgets": {
        "fields": ",".join(
            [
                "id",
                "name",
                "ref",
                "created_at",
                "updated_at",
                "title",
                "widget_type",
                "meta",
                "order",
                "activity_id",
                "parent_id",
                "progress",
                "has_subwidgets",
                "scope_id",
            ]
        )
    },
    "widget": {
        "fields": ",".join(
            [
                "id",
                "name",
                "ref",
                "created_at",
                "updated_at",
                "title",
                "widget_type",
                "meta",
                "order",
                "activity_id",
                "parent_id",
                "progress",
                "has_subwidgets",
                "scope_id",
            ]
        )
    },
    "notifications": {
        "fields": ",".join(
            ["id", "subject", "status", "message", "team", "created_at", "options", "updated_at"]
        )
    },
    "notification": {
        "fields": ",".join(
            ["id", "subject", "status", "message", "team", "created_at", "options", "updated_at"]
        )
    },
    "expiring_downloads": {
        "fields": ",".join(
            ["id", "created_at", "updated_at", "expires_at", "expires_in", "content"]
        )
    },
    "context": {
        "fields": ",".join(
            [
                "id",
                "name",
                "ref",
                "created_at",
                "updated_at",
                "description",
                "tags",
                "context_type",
                "activities",
                "scope",
                "options",
                "feature_collection",
                "start_date",
                "due_date",
            ]
        )
    },
    "contexts": {
        "fields": ",".join(
            [
                "id",
                "name",
                "ref",
                "created_at",
                "updated_at",
                "description",
                "tags",
                "context_type",
                "activities",
                "scope",
                "options",
                "feature_collection",
                "start_date",
                "due_date",
            ]
        )
    },
}
